#!/usr/bin/en"v python3
from PyQt5 import QtCore, QtGui, QtWidgets, QtSerialPort
from PyQt5.QtCore import QObject
from PyQt5.QtSerialPort import QSerialPort

from msgtools.lib.messaging import Messaging
from msgtools.lib.header_translator import HeaderTranslator, HeaderHelper
from msgtools.server import SerialportDialog

import sys

class BaseSerialConnection(QObject):
    statusUpdate = QtCore.pyqtSignal(str)
    messagereceived = QtCore.pyqtSignal(object)
    disconnected = QtCore.pyqtSignal(object)
    def __init__(self, name, portName, baud_rate=QSerialPort.Baud115200, parity=QSerialPort.NoParity):
        super(BaseSerialConnection, self).__init__(None)
        
        self.base_name = name
        self.settings = QtCore.QSettings("MsgTools", "MessageServer/"+self.base_name)
        
        self.removeClient = QtWidgets.QPushButton("Remove")
        self.removeClient.pressed.connect(self.onDisconnected)

        # button to open/close serial port
        self.openCloseButton = QtWidgets.QPushButton("button")
        self.openCloseButton.pressed.connect(self.openCloseSwitch)
        
        # button select new serial port
        self.selectPortButton = QtWidgets.QPushButton("Select Port")
        self.selectPortButton.pressed.connect(self.selectPort)

        self.statusLabel = QtWidgets.QLabel()
        self.rxMsgCount = 0
        self.subscriptions = {}
        self.subMask = 0
        self.subValue = 0
        self.isHardwareLink = True
        
        # if port not specified, default to last used port
        if not portName:
            portName = self.settings.value("portName", None)

        self.serialPort = QSerialPort(portName)
        self.serialPort.setBaudRate(baud_rate)
        self.serialPort.setFlowControl(QSerialPort.NoFlowControl)
        self.serialPort.setParity(parity)
        self.serialPort.setDataBits(QSerialPort.Data8)
        self.serialPort.setStopBits(QSerialPort.OneStop)
        self.serialPort.errorOccurred.connect(self.onSerialError)

        self.serialPort.readyRead.connect(self.onReadyRead)
        self.name = self.base_name + " " + self.serialPort.portName()
        self.statusLabel.setText(self.name)
        
        self.reopenTimer = QtCore.QTimer(self)
        self.reopenTimer.setInterval(1000)
        self.reopenTimer.timeout.connect(self._openSerialPort)

    def onDisconnected(self):
        self.disconnected.emit(self)
    
    def onSerialError(self, error):
        if error == QSerialPort.NoError:
            pass
        elif error == QSerialPort.ReadError:
            self.statusUpdate.emit("Error, Closing SerialPort on port "+str(self.serialPort.portName()))
            self._closeSerialPort()
            # Don't restart the timer if it's active.
            # If we did, then a steady stream of errors would keep resetting
            # the timer, and it would never go off.
            if not self.reopenTimer.isActive():
                self.reopenTimer.start()
        elif error == QSerialPort.ResourceError:
            pass

    def widget(self, index):
        if index == 0:
            return self.removeClient
        if index == 1:
            return self.openCloseButton
        if index == 2:
            return self.selectPortButton
        if index == 3:
            return self.statusLabel
        return None
    
    def _openSerialPort(self):
        if self.serialPort.isOpen():
            # if we're already open, make sure we stop retrying to open
            self.reopenTimer.stop()
        else:
            if self.serialPort.open(QSerialPort.ReadWrite):
                self.statusUpdate.emit("Opened SerialPort on port "+str(self.serialPort.portName()))
                self.openCloseButton.setText("Close")
                self.settings.setValue("portName", self.serialPort.portName())
                self.reopenTimer.stop()
                return True
            else:
                self.openCloseButton.setText("Open")
                return False

    def _closeSerialPort(self):
        self.serialPort.close()
        self.openCloseButton.setText("Open")

    def openCloseSwitch(self):
        # open or close the port
        if self.serialPort.isOpen():
            self._closeSerialPort()
            self.statusUpdate.emit("Closed SerialPort on port "+str(self.serialPort.portName()))
        else:
            if not self._openSerialPort():
                self.statusUpdate.emit("Can't open SerialPort on port "+str(self.serialPort.portName())+"!")

    def portChanged(self, portName):
        self.statusUpdate.emit("switching to serial port " + portName)
        if self.serialPort.isOpen():
            self.serialPort.close()
        self.serialPort.setPortName(portName)
        self.name = self.base_name + " " + self.serialPort.portName()
        self.statusLabel.setText(self.name)
        self.openCloseSwitch()

    def selectPort(self):
        d = SerialportDialog.SelectSerialportDialog()
        d.portChanged.connect(self.portChanged)
        d.exec_()

    def start(self):
        self.openCloseSwitch()

    def print_error(self, errType):
        print(errType)
        sys.stdout.flush()

    def stop(self):
        pass


class SerialConnection(BaseSerialConnection):
    def __init__(self, hdr, portName):
        super(SerialConnection, self).__init__("Serial", portName)

        self.hdr = hdr
        
        self.hdrTranslator = HeaderTranslator(hdr, Messaging.hdr)
        self.hdrHelper = HeaderHelper(hdr, self.print_error)

        self.tmpRxHdr = None

    def onReadyRead(self):
        while self.serialPort.bytesAvailable() > 0:
            if not self.tmpRxHdr:
                if self.serialPort.bytesAvailable() < self.hdr.SIZE:
                    return
                while self.serialPort.bytesAvailable() >= self.hdr.SIZE:
                    rx_bytes = bytes(self.serialPort.peek(self.hdr.SIZE))
                    serialHdr = self.hdr(rx_bytes)
                    if self.hdrHelper.header_valid(serialHdr):
                        # read now, because we only peeked before.
                        self.tmpRxHdr = self.hdr(self.serialPort.read(self.hdr.SIZE))
                        break
                    else:
                        throwAway = self.serialPort.read(1)

            if self.tmpRxHdr != None:
                bodylen = self.tmpRxHdr.GetDataLength()
                if self.serialPort.bytesAvailable() >= bodylen:
                    # allocate the serial message body, read from the serial port
                    msgBody = self.serialPort.read(bodylen)
                    
                    if self.hdrHelper.body_valid(self.tmpRxHdr, msgBody):
                        self.rxMsgCount+=1
                        self.SerialMsgSlot(self.tmpRxHdr, msgBody)
                    self.tmpRxHdr = None
                else:
                    break

    def SerialMsgSlot(self, serialHdr, body):
        networkMsg = self.hdrTranslator.translateHdrAndBody(serialHdr, body)
        self.messagereceived.emit(networkMsg)

    def sendMsg(self, networkMsg):
        serialMsg = self.hdrTranslator.translate(networkMsg)

        # if we can't translate the message, just return
        if serialMsg == None:
            return

        self.hdrHelper.finalize(serialMsg)

        if self.serialPort.isOpen():
            self.serialPort.write(serialMsg.rawBuffer().raw)
    
def PluginConnection(param=None):
    from SerialHeader import SerialHeader
    return SerialConnection(SerialHeader, param)

def PluginEnabled():
    try:
        from SerialHeader import SerialHeader
        return True
    except:
        return False

def BtPluginConnection(param=None):
    from BluetoothHeader import BluetoothHeader
    return SerialConnection(BluetoothHeader, param)

def BtPluginEnabled():
    try:
        from BluetoothHeader import BluetoothHeader
        return True
    except:
        return False

import collections
PluginInfo = collections.namedtuple('PluginInfo', ['name', 'enabled', 'connect_function'])
plugin_info = PluginInfo('Serial', PluginEnabled, PluginConnection)
bt_plugin_info = PluginInfo('Bluetooth Serial', BtPluginEnabled, BtPluginConnection)



# Used to display fields of an unknown message
import struct
import ctypes
from msgtools.lib.messaging import *
import msgtools.lib.messaging as msg
from msgtools.lib.message import Message

class UnknownMsg(Message):
    MSG_OFFSET = Messaging.hdrSize
    
    def __init__(self, messageBuffer):
        # this is a trick to get us to store a copy of a pointer to a buffer, rather than making a copy of the buffer
        self.msg_buffer_wrapper = { "msg_buffer": messageBuffer }

        self.hdr = Messaging.hdr(messageBuffer)

    def rawBuffer(self):
        # this is a trick to get us to store a copy of a pointer to a buffer, rather than making a copy of the buffer
        return self.msg_buffer_wrapper["msg_buffer"]

    def MsgName(self):
        id = hex(self.hdr.GetMessageID())
        return "Unknown_"+id

    @msg.offset('0')
    @msg.count('64')
    @msg.size('1')
    def GetRawData(self, index):
        """"""
        value = struct.unpack_from('>B', self.rawBuffer(), UnknownMsg.MSG_OFFSET + index)[0]
        return hex(value)
        
    @msg.offset('0')
    @msg.count('64')
    @msg.size('1')
    def SetRawData(self, value, index):
        """"""
        struct.pack_into('>B', self.rawBuffer(), UnknownMsg.MSG_OFFSET + index, value)
    
    # Reflection information
    fields = [FieldInfo(name="rawData",type="int",units="",minVal="",maxVal="",description="",get=GetRawData,set=SetRawData,count=64, bitfieldInfo = [], enum = [])]



import sys
import datetime
import inspect
import argparse
import json

from PyQt5 import QtGui, QtWidgets, QtCore, QtNetwork

from .app import *

import msgtools.lib.msgcsv as msgcsv
import msgtools.lib.txtreewidget as txtreewidget

# Tree widget item with support for sorting by a column, and support for per-column editability
# For per-column editability to work, you need to setItemDelegate(NoEditDelegate), and
# NoEditDelegate is defined in txtreewidget.py
class TreeWidgetItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, stringList):
        QtWidgets.QTreeWidgetItem.__init__(self, stringList)
        self.editable = False
        self.editable_columns = {}

    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()
        try:
            return float(self.text(column)) < float(otherItem.text(column))
        except ValueError:
            return self.text(column) < otherItem.text(column)

    def makeEditable(self, editable_columns):
        self.editable = True
        self.setFlags(self.flags() | QtCore.Qt.ItemIsEditable)
        for c in editable_columns:
            self.editable_columns[c] = True

    def editableColumn(self, column):
        if not self.editable:
            return False
        if column in self.editable_columns:
            return True
        return False

# tree widget that we can copy-and-paste on
class TreeWidget(QtWidgets.QTreeWidget):
    def __init__(self):
        super(TreeWidget, self).__init__() # what constructor parameters?!
        self.setSelectionMode(self.ContiguousSelection)
        self.setSelectionBehavior(self.SelectRows)

    def keyPressEvent(self, keyEvent):
        if keyEvent.matches(QtGui.QKeySequence.Copy):
            self.copySelection()
        else:
            super(TreeWidget, self).keyPressEvent(keyEvent)

    def copySelection(self):
        selectedItems = self.selectedItems()
        copiedText = ""
        h = self.headerItem()
        for col in range(h.columnCount()):
            if col != 0:
                copiedText += ", "
            copiedText += h.text(col)
        copiedText += "\n"
        
        for itemNumber in range(len(selectedItems)):
            w = selectedItems[itemNumber]
            for col in range(w.columnCount()):
                # treat first column as time, convert HH:MM:SS.sss to a float
                if col == 0:
                    t = w.text(col)
                    if ":" in t:
                        time_components = t.split(':')
                        t = int(time_components[0]) * 3600 + int(time_components[1]) * 60 + float(time_components[2])
                        copiedText += str(t)
                    else:
                        copiedText += str(t)
                else:
                    copiedText += ", "
                    copiedText += w.text(col)
            copiedText += "\n"
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(copiedText)

# text widget that allows us to push msg data directly into the text fields
class MsgTextWidget(QtWidgets.QWidget):
    def __init__(self, msgClass, allowedFields=[]):
        super(MsgTextWidget, self).__init__()
        self.grid = QtWidgets.QGridLayout()
        self.setLayout(self.grid)
        from msgtools.lib.unknownmsg import UnknownMsg
        if msgClass == UnknownMsg:
            self.showHeader = True
        else:
            self.showHeader = False
        self.allowedFields = dict((fieldName,True) for fieldName in allowedFields)
        tableHeader = []
        if self.fieldAllowed("Time"):
            timeUnits = Messaging.findFieldInfo(Messaging.hdr.fields, "Time").units
            if timeUnits == "ms":
                timeUnits = "s"
            tableHeader.append("Time ("+timeUnits+")")
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    if self.fieldAllowed(fieldInfo.name):
                        tableHeader.append(fieldInfo.name)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        if self.fieldAllowed(bitInfo.name):
                            tableHeader.append(bitInfo.name)
        for fieldInfo in msgClass.fields:
            if self.fieldAllowed(fieldInfo.name):
                tableHeader.append(fieldInfo.name)
            for bitInfo in fieldInfo.bitfieldInfo:
                if self.fieldAllowed(bitInfo.name):
                    tableHeader.append(bitInfo.name)
        self.setHeaderLabels(tableHeader)
    
    def fieldAllowed(self, fieldName):
        if len(self.allowedFields) == 0:
            return True
        return fieldName in self.allowedFields

    def setHeaderLabels(self, headerLabels):
        i = 0
        for h in headerLabels:
            self.grid.addWidget(QtWidgets.QLabel(h), 0, i)
            self.grid.addWidget(QtWidgets.QLabel(''), 1, i)
            i += 1
    
    def setDataLabels(self, dataLabels, dataAlerts):
        i = 0
        for t in dataLabels:
            self.grid.itemAtPosition(1,i).widget().setText(t)
            if dataAlerts[i]:
                #TODO Change text color
                pass
            i += 1

    def addData(self, msg, autoscroll=True):
        msgStringList = []
        columnAlerts = []
        if self.fieldAllowed("Time"):
            try:
                timeVal = msg.hdr.GetTime()
                timeInfo = Messaging.findFieldInfo(msg.hdr.fields, "Time")
                if timeInfo.units == "ms":
                    timeVal = timeVal / 1000.0
                timeVal = datetime.datetime.fromtimestamp(timeVal, datetime.timezone.utc)
            except AttributeError:
                timeVal = datetime.datetime.now()
            timeVal = timeVal.strftime('%H:%M:%S.%f')[:-3]
            msgStringList.append(timeVal)
            columnAlerts.append(0)
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    if self.fieldAllowed(fieldInfo.name):
                        fieldValue = str(Messaging.get(msg.hdr, fieldInfo))
                        msgStringList.append(fieldValue)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        if self.fieldAllowed(bitInfo.name):
                            fieldValue = str(Messaging.get(msg.hdr, bitInfo))
                            msgStringList.append(fieldValue)
        for fieldInfo in type(msg).fields:
            if(fieldInfo.count == 1):
                if self.fieldAllowed(fieldInfo.name):
                    fieldValue = str(Messaging.get(msg, fieldInfo))
                    msgStringList.append(fieldValue)
                    columnAlerts.append(Messaging.getAlert(msg, fieldInfo))
                for bitInfo in fieldInfo.bitfieldInfo:
                    if self.fieldAllowed(bitInfo.name):
                        fieldValue = str(Messaging.get(msg, bitInfo))
                        msgStringList.append(fieldValue)
                        columnAlerts.append(Messaging.getAlert(msg, bitInfo))
            else:
                if self.fieldAllowed(fieldInfo.name):
                    columnText = ""
                    alert = False
                    for i in range(0,fieldInfo.count):
                        fieldValue = str(Messaging.get(msg, fieldInfo, i))
                        # if the value is what is given when we go off the end of an array, break.
                        if fieldInfo.type == "int" and fieldValue == "UNALLOCATED":
                            break
                        columnText += str(fieldValue)
                        if Messaging.getAlert(msg, fieldInfo, i):
                            alert = True
                        if(i<fieldInfo.count-1):
                            columnText += ", "
                    msgStringList.append(columnText)
                    columnAlerts.append(alert)

        self.setDataLabels(msgStringList, columnAlerts)

# tree widget that allows us to push msg data directly into the tree
class MsgTreeWidget(TreeWidget):
    ROWS_TO_DELETE = 50
    ARRAY_STRING_MAX_LENGTH = 64
    def __init__(self, msgClass, keyField=None, maxRows=1000, rowsToDelete=50):
        super(MsgTreeWidget, self).__init__()
        from msgtools.lib.unknownmsg import UnknownMsg
        if msgClass == UnknownMsg:
            self.showHeader = True
        else:
            self.showHeader = False
        # add table header, one column for each message field
        tableHeader = []
        timeUnits = Messaging.findFieldInfo(Messaging.hdr.fields, "Time").units
        if timeUnits == "ms":
            timeUnits = "s"
        tableHeader.append("Time ("+timeUnits+")")
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    tableHeader.append(fieldInfo.name)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        tableHeader.append(bitInfo.name)
        for fieldInfo in msgClass.fields:
            tableHeader.append(fieldInfo.name)
            for bitInfo in fieldInfo.bitfieldInfo:
                tableHeader.append(bitInfo.name)
        self.setHeaderLabels(tableHeader)
        # configure the header so we can click on it to sort
        self.header().setSectionsClickable(1)
        self.header().setSortIndicatorShown(1)
        # show sort indicator ascending on Time, if not sorting, because we append incoming messages
        self.header().setSortIndicator(0, QtCore.Qt.AscendingOrder)
        self.header().sectionClicked.connect(self.tableHeaderClicked)

        # Use a timer to autoscroll, to reduce CPU usage and make it a little
        # easier to read the screen while it's scrolling.
        self.scrollTimer = QtCore.QTimer()
        self.scrollTimer.setInterval(500)
        self.scrollTimer.setSingleShot(True)
        self.scrollTimer.timeout.connect(self.scrollToBottom)

        self.maxRows = maxRows
        if rowsToDelete > maxRows:
            rowsToDelete = maxRows
        self.rowsToDelete = rowsToDelete

        # key field for our message, that we display one row per value of
        self.keyField = keyField

        self.firstTimeDataAdded = True

    def addData(self, msg, autoscroll=True):
        msgStringList = []
        columnAlerts = []
        try:
            timeVal = msg.hdr.GetTime()
            timeInfo = Messaging.findFieldInfo(msg.hdr.fields, "Time")
            if timeInfo.units == "ms":
                timeVal = timeVal / 1000.0
            timeVal = datetime.datetime.fromtimestamp(timeVal, datetime.timezone.utc)
        except AttributeError:
            timeVal = datetime.datetime.now()
        timeVal = timeVal.strftime('%H:%M:%S.%f')[:-3]
        msgStringList.append(timeVal)
        columnAlerts.append(0)
        keyColumn = -1
        columnCounter = 1
        if self.showHeader:
            for fieldInfo in Messaging.hdr.fields:
                if len(fieldInfo.bitfieldInfo) == 0:
                    fieldValue = str(Messaging.get(msg.hdr, fieldInfo))
                    msgStringList.append(fieldValue)
                    columnCounter += 1
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        fieldValue = str(Messaging.get(msg.hdr, bitInfo))
                        msgStringList.append(fieldValue)
                        columnCounter += 1
        for fieldInfo in type(msg).fields:
            if(fieldInfo.count == 1):
                fieldValue = str(Messaging.get(msg, fieldInfo))
                msgStringList.append(fieldValue)
                columnAlerts.append(Messaging.getAlert(msg, fieldInfo))
                if fieldInfo.name == self.keyField:
                    keyValue = fieldValue
                    keyColumn = columnCounter
                columnCounter += 1
                for bitInfo in fieldInfo.bitfieldInfo:
                    fieldValue = str(Messaging.get(msg, bitInfo))
                    msgStringList.append(fieldValue)
                    columnAlerts.append(Messaging.getAlert(msg, bitInfo))
                    if bitInfo.name == self.keyField:
                        keyValue = fieldValue
                        keyColumn = columnCounter
                    columnCounter += 1
            else:
                columnText = ""
                alert = False
                if "hex" == fieldInfo.units.lower():
                    columnText = "0x"
                    for i in range(0, fieldInfo.count):
                        fieldValue = Messaging.get(msg, fieldInfo, i).replace("0x","")
                        if fieldInfo.type == "int" and fieldValue == "UNALLOCATED":
                            break
                        columnText += fieldValue
                        if len(columnText) > MsgTreeWidget.ARRAY_STRING_MAX_LENGTH:
                            columnText += "..."
                            break
                else:
                    for i in range(0,fieldInfo.count):
                        fieldValue = str(Messaging.get(msg, fieldInfo, i))
                        # if the value is what is given when we go off the end of an array, break.
                        if fieldInfo.type == "int" and fieldValue == "UNALLOCATED":
                            break
                        columnText += str(fieldValue)
                        if Messaging.getAlert(msg, fieldInfo, i):
                            alert = True
                        if(i<fieldInfo.count-1):
                            columnText += ", "
                msgStringList.append(columnText)
                columnAlerts.append(alert)
                columnCounter += 1

        msgItem = TreeWidgetItem(msgStringList)
        for column in range(0, len(columnAlerts)):
            if columnAlerts[column]:
                font = msgItem.font(column)
                brush = msgItem.foreground(column)
                font.setBold(1)
                brush.setColor(QtCore.Qt.red)
                msgItem.setFont(column, font)
                msgItem.setForeground(column, brush)
                msgItem.setBackground(column, brush)
        if self.keyField != None and keyColumn >= 0:
            # find row that has key field that matches ours
            foundAndReplaced = False
            for i in range(0, self.topLevelItemCount()):
                item = self.topLevelItem(i)
                if item.text(keyColumn) == keyValue:
                    foundAndReplaced = True
                    self.takeTopLevelItem(i)
                    self.insertTopLevelItem(i, msgItem)
            if not foundAndReplaced:
                self.addTopLevelItem(msgItem)
                self.sortItems(keyColumn, QtCore.Qt.AscendingOrder)
        else:
            self.addTopLevelItem(msgItem)
            if autoscroll and not self.scrollTimer.isActive():
                self.scrollTimer.start()

            # when deleting, make a bunch of room so that if we're auto scrolling we have a bit of time before it
            # shifts the data.  Otherwise the user can't read stuff if it's going by too fast.
            if self.topLevelItemCount() > self.maxRows:
                for i in range(0, self.rowsToDelete):
                    self.takeTopLevelItem(0)
                    
        if self.firstTimeDataAdded:
            count = 0
            for fieldInfo in type(msg).fields:
                self.resizeColumnToContents(count)
                count += 1
        self.firstTimeDataAdded = False
        
    def tableHeaderClicked(self, column):
        fieldName = self.headerItem().text(column)
        if self.keyField == None or self.keyField != fieldName:
            self.keyField = fieldName
            self.header().setSortIndicator(column, QtCore.Qt.AscendingOrder)
            self.sortItems(column, QtCore.Qt.AscendingOrder)
            valueToRemove = None
            for i in range(self.topLevelItemCount()-1, -1 ,-1):
                item = self.topLevelItem(i)
                if not valueToRemove == None and item.text(column) == valueToRemove:
                    self.takeTopLevelItem(i)
                else:
                    valueToRemove = item.text(column)
        else:
            self.keyField = None
            self.header().setSortIndicator(0, QtCore.Qt.AscendingOrder)

class LineEditWithHistory(QtWidgets.QLineEdit):
    tabPressed = QtCore.pyqtSignal()
    def __init__(self, parent=None):
        super(LineEditWithHistory, self).__init__()
        self.commandHistory = []
        self.placeInHistory = 0

    def setToHistoryItem(self, index):
        self.placeInHistory = index
        if self.placeInHistory > len(self.commandHistory) - 1:
            self.placeInHistory = len(self.commandHistory)
            self.setText('')
        else:
            if self.placeInHistory < 0:
                self.placeInHistory = 0
            try:
                self.setText(self.commandHistory[self.placeInHistory])
            except IndexError:
                self.setText('')
    
    # disable tab focus so we get tab keys delivered via keyPressEvent.
    def focusNextPrevChild(self, next):
        return False
    
    def addToHistory(self, lineOfText):
        if len(self.commandHistory) == 0 or self.commandHistory[-1] != lineOfText:
            self.commandHistory.append(lineOfText)
            if len(self.commandHistory) > 20:
                self.commandHistory.pop(0)
        self.placeInHistory = len(self.commandHistory)

    def keyPressEvent(self, keyEvent):
        if keyEvent.key() == QtCore.Qt.Key_Return:
            # add to history
            self.addToHistory(self.text())
        elif keyEvent.key() == QtCore.Qt.Key_Up:
            #up in history
            self.setToHistoryItem(self.placeInHistory - 1)
        elif keyEvent.key() == QtCore.Qt.Key_Down:
            # down in history
            self.setToHistoryItem(self.placeInHistory + 1)
        elif keyEvent.key() == QtCore.Qt.Key_Tab:
            self.tabPressed.emit()
            return
        super(LineEditWithHistory, self).keyPressEvent(keyEvent)

class MsgCommandWidget(QtWidgets.QWidget):
    commandEntered = QtCore.pyqtSignal(str)
    messageEntered = QtCore.pyqtSignal(object)
    autocompleted   = QtCore.pyqtSignal(str)
    def __init__(self, parent=None):
        super(MsgCommandWidget, self).__init__()
        self.textBox = QtWidgets.QPlainTextEdit()
        self.textBox.setReadOnly(True)
        self.lastHtml = False
        self.lineEdit = LineEditWithHistory()
        self.lineEdit.returnPressed.connect(self.returnPressed)
        self.lineEdit.tabPressed.connect(self.tabPressed)
        
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(self.textBox)
        vbox.addWidget(self.lineEdit)
        self.setLayout(vbox)
    
    def tabPressed(self):
        lineOfText = self.lineEdit.text()
        autocomplete, help = msgcsv.csvHelp(lineOfText)
        if autocomplete:
            self.lineEdit.setText(autocomplete)
            self.autocompleted.emit(autocomplete)
        if help:
            self.addText('\n'+help+'\n>\n')

    def returnPressed(self):
        lineOfText = self.lineEdit.text()
        self.addText(lineOfText)
        msg = msgcsv.csvToMsg(lineOfText)
        if msg:
            self.messageEntered.emit(msg)
            self.addText(" -> Msg\n")
        else:
            self.commandEntered.emit(lineOfText)
        self.lineEdit.setText("")

    def addText(self, text, errorKnown=0):
        lines = text.splitlines()
        for i in range(len(lines)):
            line = lines[i]
            if i < len(lines)-1:
                line = line + '\n'
            elif i == len(lines)-1 and text.endswith('\n'):
                line = line + '\n'
            self.addLine(line, errorKnown)
    
    def addLine(self, line, errorKnown):
        if errorKnown == 2 or 'error' in line.lower() or 'fail' in line.lower():
            self.textBox.appendHtml('<font color="red">'+line+'</font>')
            self.lastHtml = True
        elif errorKnown == 1 or 'warning' in line.lower():
            self.textBox.appendHtml('<font color="orange">'+line+'</font>')
            self.lastHtml = True
        else:
            if self.lastHtml:
                self.textBox.appendHtml('<font color="black"> </font>')
            self.textBox.moveCursor (QtGui.QTextCursor.End)
            self.textBox.insertPlainText(line)
            self.textBox.moveCursor (QtGui.QTextCursor.End)
            self.lastHtml = False

    def clear(self):
        self.textBox.clear()
    
    def addToHistory(self, lineOfText):
        self.lineEdit.addToHistory(lineOfText)
    
    def saveState(self):
        ret = ""
        for h in self.lineEdit.commandHistory:
            ret = ret + "|" + h
        return ret
    
    def restoreState(self, state):
        if state:
            history = state.split("|")
            for h in history:
                self.lineEdit.addToHistory(h)

class Gui(App, QtWidgets.QMainWindow):
    @classmethod
    def addBaseArguments(cls, parser):
        '''
        Adds base app arguments to the provided ArgParser

        returns the parser
        '''

        # Just delegate - this design choice is two fold - it prevents
        # code using the Gui from including the App as a dependency.
        # It also allows us to inject Gui specific arguments into this
        # base class later.
        return App.addBaseArguments(parser)

    '''Gui base class that provides standard connection and status UI

        name - the name of the app - will be displayed in the title bar of the window
        args - argparse.Namespace representing parsed arguments.  See app.py for
               the standard options provided as part of this Gui/App framework used
               in most MsgTools utilities.
        parent - a parent Qt Widget to attach to
    '''
    def __init__(self, name, args, parent=None):

        QtWidgets.QMainWindow.__init__(self,parent)
        App.__init__(self, name, args)

        # make a status bar to print status messages
        self.status = QtWidgets.QLabel("Initializing")
        self.statusBar().addPermanentWidget(self.status)
        # hook it up to our base class statusUpdate signal
        self.statusUpdate.connect(self.status.setText)
        
        # a checkbox for connection state to the status bar
        self.connectionCheckbox = QtWidgets.QCheckBox("")
        self.statusBar().addPermanentWidget(self.connectionCheckbox)
        # hook it up to connect/disconnect
        self.connectionCheckbox.clicked.connect(self.connectionCheckedChanged)
        self.connectionChanged.connect(self.connectionChangedSlot)

        QtCore.QTimer.singleShot(0, self.delayedInit)
        self.setWindowTitle(self.name)

        # create menu items, connect them to socket operations
        if(self.connectionType.lower() == "socket" or self.connectionType.lower() == "qtsocket"):
            connectAction = QtWidgets.QAction('&Connect', self)
            disconnectAction = QtWidgets.QAction('&Disconnect', self)

            menubar = self.menuBar()
            connectMenu = menubar.addMenu('&Connect')
            connectMenu.addAction(connectAction)
            connectMenu.addAction(disconnectAction)
            connectAction.triggered.connect(self.chooseHost)
            disconnectAction.triggered.connect(self.CloseConnection)
    
    def delayedInit(self):
        if hasattr(self.__class__, 'on_open'):
            self.on_open()
        self.readSettings()

    # open dialog box to choose host to connect to
    def chooseHost(self):
        userInput, ok = QInputDialog.getText(self, 'Connect',  'Server:', QLineEdit.Normal, self.connectionName)
        if ok:
            self.connectionName = userInput
            self.OpenConnection()
    
    def closeEvent(self, event):
        self.settings.setValue("geometry", self.saveGeometry())
        self.settings.setValue("windowState", self.saveState())
        if hasattr(self.__class__, 'on_close'):
            self.on_close()
        super(Gui, self).closeEvent(event)
    
    def readSettings(self):
        self.restoreGeometry(self.settings.value("geometry", QtCore.QByteArray()))
        self.restoreState(self.settings.value("windowState", QtCore.QByteArray()))

    def connectionChangedSlot(self, connected):
        self.connectionCheckbox.setChecked(connected)
        if connected:
            self.statusUpdate.emit('')
            self.connectionCheckbox.setText("Connected")
        else:
            self.connectionCheckbox.setText("NOT Connected")

    def connectionCheckedChanged(self, checked):
        self.statusUpdate.emit('')
        if checked:
            self.connectionCheckbox.setText("Connecting")
            self.OpenConnection()
        else:
            self.connectionCheckbox.setText("NOT Connected")
            self.CloseConnection()

import re
def get_trailing_number(s):
    m = re.search(r'\d+$', s)
    return int(m.group()) if m else None

def rchop(s, suffix):
    if suffix and s.endswith(suffix):
        return s[:-len(suffix)]
    return s

def get_type(s):
    type_names = {"_f":"float", "_u":"uint32", "_i":"int32"}
    if s.endswith(tuple(type_names.keys())):
        return type_names[s[-2:]]
    else:
        return ""

def get_name_type_and_count(s):
    count = get_trailing_number(s)
    if count != None:
        s = rchop(s, str(count))
    else:
        count = 1
    type = get_type(s)
    name = s[:-2]
    return name, type, count

class ConfigTreeWidgetItem(TreeWidgetItem):
    NAME_COLUMN = 0
    TYPE_COLUMN = 1
    DEVICE_VALUE_COLUMN = 2
    SYNC_COLUMN = 3
    FILE_VALUE_COLUMN = 4
    COLUMN_COUNT = 5
    def __init__(self, key):
        name, type, count = get_name_type_and_count(key)
        if count == 1:
            type_string = type
        else:
            type_string = "%s[%s]" % (type, count)
        string_list = [name, type_string, "", "", ""]
        super().__init__(string_list)
        self.makeEditable([ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN, ConfigTreeWidgetItem.FILE_VALUE_COLUMN])
        
        self.key = key
        self.name = name
        self.type = type
        self.count = count
        
        self.device_value = None
        self.file_value = None
        sync_left = QtWidgets.QToolButton()
        sync_left.setText("\u2190")
        sync_right = QtWidgets.QToolButton()
        sync_right.setText("\u2192")
        self.sync_widget = QtWidgets.QWidget()
        self.sync_widget.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(sync_left)
        hbox.addWidget(sync_right)
        self.sync_widget.setLayout(hbox)
        sync_left.clicked.connect(self.syncLeft)
        sync_right.clicked.connect(self.syncRight)

    def data(self, column, role):
        if role != QtCore.Qt.FontRole and role != QtCore.Qt.ForegroundRole:
            return super(ConfigTreeWidgetItem, self).data(column, role)

        if column == ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN:
            alert = self.device_value != self.text(ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN)
            alert_color = QtGui.QColor("darkOrange")
        elif column == ConfigTreeWidgetItem.FILE_VALUE_COLUMN:
            alert = self.file_value != self.text(ConfigTreeWidgetItem.FILE_VALUE_COLUMN)
            alert_color = QtGui.QColor("darkOrange")
        else:
            device_value = self.text(ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN)
            file_value = self.text(ConfigTreeWidgetItem.FILE_VALUE_COLUMN)
            alert = False if device_value == file_value else True
            alert_color = QtCore.Qt.red
        if alert:
            self.treeWidget().alert.emit(alert_color)
        if role == QtCore.Qt.FontRole:
            font = QtGui.QFont()
            if alert:
                font.setBold(True)
            return font
        if role == QtCore.Qt.ForegroundRole:
            brush = QtGui.QBrush()
            if alert:
                brush.setColor(alert_color)
            return brush

        return super(ConfigTreeWidgetItem, self).data(column, role)

    def syncLeft(self):
        self.setText(ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN, self.text(ConfigTreeWidgetItem.FILE_VALUE_COLUMN))

    def syncRight(self):
        self.setText(ConfigTreeWidgetItem.FILE_VALUE_COLUMN, self.text(ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN))
    
    def setDeviceValue(self, value):
        self.device_value = value
        self.setText(ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN, value)

    def setFileValue(self, value):
        self.file_value = value
        self.setText(ConfigTreeWidgetItem.FILE_VALUE_COLUMN, value)

# Tree widget that synchronizes configuration settings between a device and a config file,
# and lets the user edit both.
class ConfigTreeWidget(TreeWidget):
    send_msg = QtCore.pyqtSignal(object)
    alert = QtCore.pyqtSignal(object)
    ARRAY_STRING_MAX_LENGTH = 64
    def __init__(self, msg_namespace, cfg_filename):
        super(ConfigTreeWidget, self).__init__()
        self.msg_namespace = msg_namespace
        self.cfg_filename = cfg_filename

        # add table header, one column for each message field
        tableHeader = []
        tableHeader.append("Setting Name")
        tableHeader.append("Type")
        tableHeader.append("Device Value")
        tableHeader.append("Sync")
        tableHeader.append("File Value")
        self.setHeaderLabels(tableHeader)
        
        self.items_by_key = {}
        for key, value in self.msg_namespace.GetConfigValue.ConfigSettingsKeys.items():
            # Don't use value 0, it's a marker for an Invalid value.
            if value == 0:
                continue
            tree_item = ConfigTreeWidgetItem(key)
            self.addTopLevelItem(tree_item)
            self.setItemDelegate(txtreewidget.NoEditDelegate(self, self))
            self.items_by_key[key] = tree_item
            self.setItemWidget(tree_item, ConfigTreeWidgetItem.SYNC_COLUMN, tree_item.sync_widget)

        for c in range(ConfigTreeWidgetItem.COLUMN_COUNT):
            self.resizeColumnToContents(c)
    
    def load_from_device(self):
        query_msg = self.msg_namespace.GetConfigKeys()
        self.send_msg.emit(query_msg)

    def save_to_device(self):
        for key, tree_item in self.items_by_key.items():
            values = tree_item.text(ConfigTreeWidgetItem.DEVICE_VALUE_COLUMN)
            if values == "":
                delete_value_msg = self.msg_namespace.DeleteConfigValue()
                delete_value_msg.SetKey(key)
                self.send_msg.emit(delete_value_msg)
            else:
                set_value_msg = self.msg_namespace.SetConfigValue()
                set_value_msg.SetKey(key)
                idx = 0
                for value in values.split(","):
                    set_value_msg.SetValues(float(value), idx)
                    idx += 1
                set_value_msg.SetCount(idx)
                self.send_msg.emit(set_value_msg)
        # After saving to device, re-load from the device
        self.load_from_device()

    def process_msg(self, msg):
        if type(msg) == self.msg_namespace.CurrentConfigKeys:
            # if we got a message with a list of config setting keys,
            # iterate through the keys and request the config setting for each.
            for i in range(msg.GetCount()):
                query_msg = self.msg_namespace.GetConfigValue()
                query_msg.SetKey(msg.GetKey(i))
                self.send_msg.emit(query_msg)

        elif type(msg) == self.msg_namespace.CurrentConfigValue:
            # if we got a message with the value of a config setting,
            # display it in the table, unless it's an invalid key
            key_int = msg.GetKey(enumAsInt=True)
            key = msg.GetKey()
            if key == "Invalid" or key_int == 0:
                return
            if key in self.items_by_key:
                count = msg.GetCount()
                value = ""
                for i in range(count):
                    value += str(msg.GetValues(i)) + ", "
                value = value[:-2]
                tree_item = self.items_by_key[key]
                tree_item.setDeviceValue(value)
                self.viewport().update()
            else:
                print("Key %s not in keys %s" % (key, list(self.items_by_key.keys())))

    def load_from_file(self):
        with open(self.cfg_filename, 'r') as f:
            for line in f:
                data = json.loads(line)
                for key, value in data.items():
                    try:
                        tree_item = self.items_by_key[key]
                        tree_item.setFileValue(value)
                    except KeyError:
                        print("ERROR! Invalid Key %s not in %s" % (key, self.items_by_key.keys()))
        self.viewport().update()

    def save_to_file(self):
        with open(self.cfg_filename, 'w') as f:
            for key, tree_item in self.items_by_key.items():
                value = tree_item.text(ConfigTreeWidgetItem.FILE_VALUE_COLUMN)
                if value:
                    d = {key: value}
                    f.write(json.dumps(d)+"\n")
        # After writing, re-read the file.
        self.load_from_file()

class ConfigEditor(QtWidgets.QWidget):
    send_msg = QtCore.pyqtSignal(object)
    alert = QtCore.pyqtSignal(object)
    def __init__(self, msg_namespace, cfg_filename):
        super(ConfigEditor, self).__init__()

        hbox = QtWidgets.QHBoxLayout()
        load_from_device_button = QtWidgets.QPushButton("Load from device")
        save_to_device_button = QtWidgets.QPushButton("Save to device")
        load_from_file_button = QtWidgets.QPushButton("Load from file")
        save_to_file_button = QtWidgets.QPushButton("Save to file")
        hbox.addWidget(load_from_device_button)
        hbox.addWidget(save_to_device_button)
        hbox.addWidget(load_from_file_button)
        hbox.addWidget(save_to_file_button)

        self.edit_tree = ConfigTreeWidget(msg_namespace, cfg_filename)
        self.edit_tree.send_msg.connect(self.send_msg)
        self.edit_tree.alert.connect(self.alert)

        load_from_file_button.clicked.connect(self.edit_tree.load_from_file)
        load_from_device_button.clicked.connect(self.edit_tree.load_from_device)
        save_to_file_button.clicked.connect(self.edit_tree.save_to_file)
        save_to_device_button.clicked.connect(self.edit_tree.save_to_device)
        
        vbox = QtWidgets.QVBoxLayout()
        vbox.addLayout(hbox)
        vbox.addWidget(self.edit_tree)
        self.setLayout(vbox)
    
    def process_msg(self, msg):
        self.edit_tree.process_msg(msg)
    


#    <OUTPUTFILENAME>
#    Created <DATE> from:
#        Messages = <INPUTFILENAME>
#        Template = <TEMPLATEFILENAME>
#        Language = <LANGUAGEFILENAME>
#
#                     AUTOGENERATED FILE, DO NOT EDIT
import struct
import ctypes
from collections import OrderedDict
from msgtools.lib.messaging import *
import msgtools.lib.messaging as msg

class <MSGNAME> :
    SIZE = <MSGSIZE>
    MSG_OFFSET = 0
    # Enumerations
    <ENUMERATIONS>

    #@staticmethod
    #def Create() :
    #    message_buffer = ctypes.create_string_buffer(<MSGNAME>.SIZE)
    #    <INIT_CODE>
    #    return message_buffer
    
    def __init__(self, messageBuffer=None):
        doInit = 0
        if messageBuffer == None:
            doInit = 1
            messageBuffer = ctypes.create_string_buffer(<MSGNAME>.SIZE)
        else:
            try:
                messageBuffer.raw
            except AttributeError:
                newbuf = ctypes.create_string_buffer(len(messageBuffer))
                for i in range(0, len(messageBuffer)):
                    newbuf[i] = bytes(messageBuffer)[i]
                messageBuffer = newbuf
        # this is a trick to get us to store a copy of a pointer to a buffer, rather than making a copy of the buffer
        self.msg_buffer_wrapper = { "msg_buffer": messageBuffer }
        if doInit:
            self.initialize()

    def initialize(self):
            <INIT_CODE>
            pass
    
    def rawBuffer(self):
        # this is a trick to get us to store a copy of a pointer to a buffer, rather than making a copy of the buffer
        return self.msg_buffer_wrapper["msg_buffer"]

    @staticmethod
    def MsgName():
        return "<MSGNAME>"

    def SetMessageID(self, id):
        <SETMSGID>

    def GetMessageID(self):
        id = <GETMSGID>
        return id

    def __repr__(self):
        def add_param(n, v):
            if v == '':
                v = '""'
            if ',' in v and not((v.startswith('"') and v.endswith('"')) or (v.startswith("'") and v.endswith("'"))):
                v = '"%s"' % v
            return n + " = " + v + ", "
        ret = ''
        for fieldInfo in self.fields:
            if(fieldInfo.count == 1):
                if len(fieldInfo.bitfieldInfo) == 0:
                    ret += add_param(fieldInfo.name, str(Messaging.get(self, fieldInfo)))
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        ret += add_param(bitInfo.name, str(Messaging.get(self, bitInfo)))
            else:
                arrayList = []
                terminate = 0
                for i in range(0,fieldInfo.count):
                    arrayList.append(str(Messaging.get(self, fieldInfo, i)))
                ret += add_param(fieldInfo.name, "[" + ','.join(arrayList) + "]")
                if terminate:
                    break

        if ret.endswith(", "):
            ret = ret[:-2]
        return "%s(%s)" % (self.MsgName(), ret)

    # Accessors
    <ACCESSORS>

    # Reflection information
    fields = [ \
        <REFLECTION>\
    ]



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *
    
def accessors(msg):
    return []

def fieldDefault(field, as_enum=False):
    try:
        ret = field["Default"]
    except KeyError:
        ret = 0
    return ret

def enums(e):
    ret = ""
    for enum in e:
        ret +=  "enum " + enum["Name"]+" {"
        for option in enum["Options"]:
            optionName = OptionName(option)
            ret += optionName+" = "+str(option["Value"]) + ', '
        ret = ret[:-2]
        ret += "};\n"
    return ret

def cosmosType(field):
    ret = field["Type"]
    if ret.startswith("float"):
        ret = "FLOAT"
    elif ret.startswith("int"):
        ret = "INT"
    elif ret.startswith("uint"):
        ret = "UINT"
    return ret

def fieldInfos(msg):
    return ""

def scale_and_offset(field):
    scale = 1.0
    offset = 0.0
    if "Scale" in field:
        scale = field["Scale"]
    if "Offset" in field:
        offset = field["Offset"]
    return (scale, offset)

def generic_declaration(prefix, is_cmd, field, type, bit_location, bit_size, msg_enums):
    ret = ""
    min = field["Min"] if "Min" in field else "MIN"
    max = field["Max"] if "Max" in field else "MAX"
    count = MsgParser.fieldCount(field)
    array_bitsize = bit_size * count
    if is_cmd:
        if count == 1:
            ret = '%sPARAMETER %s %d %d  %s %s %s %s "%s"' % (prefix, field["Name"], bit_location, bit_size, type, min, max, fieldDefault(field), MsgParser.fieldDescription(field))
        else:
            ret = '%sARRAY_PARAMETER %s %d %d %s %s "%s"' % (prefix, field["Name"], bit_location, bit_size, type, array_bitsize, MsgParser.fieldDescription(field))
    else:
        if count == 1:
            ret = '%sITEM      %s %d %d  %s "%s"'  % (prefix, field["Name"], bit_location, bit_size, type, MsgParser.fieldDescription(field))
        else:
            ret = '%sARRAY_ITEM      %s %d %d  %s %d "%s"'  % (prefix, field["Name"], bit_location, bit_size, type, array_bitsize, MsgParser.fieldDescription(field))
        if "Min" in field or "Max" in field:
            # this is:
            # Limits Set name (DEFAULT is fine)
            # Persistence (how many violations before flagged, 1 is fine)
            # Initial State, either ENABLED or DISABLED
            # 4 required values: red low, yellow low, yellow high, red high
            # 2 optional values: green low, green high (not used)
            ret += "\n%s  LIMITS DEFAULT 1 ENABLED %s %s %s %s" % (prefix, min, min, max, max)
            
    if "Enum" in field:
        if field["Enum"] in msg_enums:
            enum = msg_enums[field["Enum"]]
            for option in enum["Options"]:
                name = OptionName(option)
                value = str(option["Value"])
                ret += "%s  STATE %s %s" % (name, value)
    if "Units" in field and field["Units"] != "":
        ret += "\n  %sUNITS %s %s" % (prefix, field["Units"].strip(), field["Units"].strip())
    if fieldHasConversion(field):
        conversion = "POLY_WRITE_CONVERSION" if is_cmd else "POLY_READ_CONVERSION"
        scale, offset = scale_and_offset(field)
        ret += "\n  %s%s %s %s" % (prefix, conversion, str(offset), str(scale))
    return ret

def declarations(msg, msg_enums):
    field_base_location = 32
    ret = []
    ret += [
        'COMMAND TARGET_NAME <MSGFULLNAME> %s "<MSGDESCRIPTION>"' % (msgEndian(msg)),
        '  ID_PARAMETER MESSAGE_ID  0  %d   UINT    <MSGID>  "Message ID"' % (field_base_location),
        '',
        '  # Fields have:        PARAMETER        Name  BitOffset BitSize Type  Min Max  Default               Description',
        '  # Array Fields have:  ARRAY_PARAMETER  Name  BitOffset BitSize Type                   ArrayBitSize  Description']
    ret += specialized_declarations(True, msg, msg_enums, field_base_location)
    ret += [
        '',
        'TELEMETRY TARGET_NAME <MSGFULLNAME> %s "<MSGDESCRIPTION>"' % (msgEndian(msg)),
        '  ID_ITEM MESSAGE_ID  0  %d   UINT    <MSGID>  "Message ID"' % (field_base_location),
        '',
        '  # Fields have:       ITEM        Name  BitOffset BitSize Type               Description',
        '  # Array Fields have: ARRAY_ITEM  Name  BitOffset BitSize Type  ArrayBitSize Description']
    ret += specialized_declarations(False, msg, msg_enums, field_base_location)
    return ret

def specialized_declarations(is_cmd, msg, msg_enums, field_base_location):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            field_bit_location = field_base_location + 8 * MsgParser.fieldLocation(field)
            ret.append(generic_declaration("  ", is_cmd, field, cosmosType(field), field_bit_location, 8*MsgParser.fieldSize(field), msg_enums))
            if "Bitfields" in field:
                ret.append("    OVERLAP")
                bitOffset = 0
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(generic_declaration("    ", is_cmd, bits, cosmosType(field), field_bit_location+bitOffset, numBits, msg_enums))
                    ret.append("      OVERLAP")
                    bitOffset += numBits
    return ret

def msgEndian(msg):
    return "BIG_ENDIAN" if MsgParser.big_endian else "LITTLE_ENDIAN"

def getMsgID(msg):
    return baseGetMsgID("", "", "CAST_ENUMS", 0, msg)
    
def setMsgID(msg):
    return baseSetMsgID("", "", 1, 0, msg)

def initCode(msg):
    return []



# for directory listing, exit function
import os, glob, sys, struct, time, json

# for runtime module importing
import importlib

# for getting current time
import datetime

# A decorator to specify units for fields
def units(arg):
    def _units(fcn):
        fcn.units = arg
        return fcn
    return _units

# A decorator to specify a default value for fields
def default(arg):
    def _default(fcn):
        fcn.default = arg
        return fcn
    return _default

# A decorator to specify a count for fields
def count(arg):
    def _count(fcn):
        fcn.count = arg
        return fcn
    return _count

# A decorator to specify a byte offset for fields
def offset(arg):
    def _offset(fcn):
        fcn.offset = arg
        return fcn
    return _offset

# A decorator to specify a size in bytes for fields
def size(arg):
    def _size(fcn):
        fcn.size = arg
        return fcn
    return _size

# A decorator to specify a min for fields
def minVal(arg):
    def _minVal(fcn):
        fcn.minVal = arg
        return fcn
    return _minVal

# A decorator to specify a max for fields
def maxVal(arg):
    def _maxVal(fcn):
        fcn.maxVal = arg
        return fcn
    return _maxVal

import collections

# This simulates a hash table, but with lazy loading, where a module isn't
# loaded until an attempt is made to use it.  Until then, only None is stored
# at each hash table location, so that keys() still returns the expected list.
class MessageNameLoader(collections.UserDict):
    def __init__(self,*arg,**kw):
        super(MessageNameLoader, self).__init__(*arg, **kw)

    def __getitem__(self, key):
        if key in self.data and self.data[key] != None:
            return self.data[key]

        if key in Messaging.MsgModuleFromName:
            importlib.import_module(Messaging.MsgModuleFromName[key])
        if not key in self.data:
            raise KeyError("Invalid Message name %s" % (key))
        return self.data[key]

# This simulates an object with attributes, but with lazy loading, where a module isn't
# loaded until an attempt is made to use it.
class MessageAttributeLoader(object):
    def __init__(self, basename):
        self.basename = basename
    
    def Find(self, msgname):
        return self.__getattr__(msgname)

    def __getattr__(self, key):
        if self.basename:
            msgname = self.basename+"."+key
        else:
            msgname = key

        if msgname in Messaging.MsgModuleFromName:
            importlib.import_module(Messaging.MsgModuleFromName[msgname])
        if key in vars(self):
            return getattr(self, key)
        raise Messaging.LoadError(self.basename, key)

class TimestampFixer:
    def __init__(self):
        self.last_rx_time = None
        self.last_rx_timestamp = None
        self.time_info = Messaging.findFieldInfo(Messaging.hdr.fields, "Time")
        TIME_SCALES = {"s": 1.0, "ms": 1000.0, "us": 1000000.0, "ns": 1e9}
        try:
            self.time_scale = TIME_SCALES[self.time_info.units]
        except KeyError:
            self.time_scale = None
    
    def restore_timestamp(self, hdr, original_timestamp):
        if original_timestamp != None:
            hdr.SetTime(original_timestamp)

    def fix_timestamp(self, hdr):
        # if there's no time field, don't do anything
        if not self.time_info:
            return None

        if hdr.GetTime() == 0.0:
            original_timestamp = hdr.GetTime()
            # if time is zero, then check if we've previously had non-zero time
            if self.last_rx_time == None:
                # if we've never had time, then just use system time, but scaled
                current_time = datetime.datetime.now().timestamp()
                if self.time_scale != None:
                    current_time *= self.time_scale
                if self.time_info.type == "int":
                    current_time = int(current_time)
                hdr.SetTime(current_time)
            else:
                # if we have had time, then try scaling between wallclock time and message time.
                # If we can scale between wallclock time and message time, compute
                # a new time based on last rx time and elapsed time.
                if self.time_scale != None:
                    current_time = datetime.datetime.now().timestamp()
                    elapsed_time = current_time - self.last_rx_time
                    elapsed_time *= self.time_scale
                    if self.time_info.type == "int":
                        elapsed_time = int(elapsed_time)
                    new_time = self.last_rx_timestamp + elapsed_time
                    hdr.SetTime(new_time)
                else:
                    # otherwise, set time to timestamp of last reception
                    hdr.SetTime(self.last_rx_timestamp)
            return original_timestamp
        else:
            # If there's a valid timestamp record it to use the next time
            # a message comes in if it has no timestamp.
            self.last_rx_time = datetime.datetime.now().timestamp()
            self.last_rx_timestamp = hdr.GetTime()
        return None

class Messaging:
    hdr=None
    hdrSize=0

    # hash tables for msg lookups
    MsgNameFromID = {}
    MsgIDFromName = {}
    MsgClassFromName = MessageNameLoader()
    MsgModuleFromName = {}

    # container for accessing message classes via attributes.
    Messages = MessageAttributeLoader("")
        
    debug=0
    objdir=None

    class LoadError(Exception):
        def __init__(self, basename, submsgname):
            if basename == "":
                fullname = "Messages"
            else:
                fullname = "Messages."+basename
            txt = "'%s' does not contain '%s.%s'" % (fullname, fullname, submsgname)
            super().__init__(txt)
            self.basename = basename
            self.submsgname = submsgname

        def __str__(self):
            error_text = super().__str__()
            suggestions = self.valid_submsgs()
            if suggestions:
                error_text += "\nSuggestions:"
                for suggestion in suggestions:
                    error_text += f"\n- {suggestion}"
            return error_text

        def valid_submsgs(self):
            ret = []
            basename = self.basename
            if basename == "":
                basename = "Messages"
            else:
                basename = "Messages."+basename
            for msgname in Messaging.MsgModuleFromName:
                fullname = "Messages."+msgname
                #print("fullname: %s" % (fullname))
                if fullname.startswith(basename):
                    if fullname.replace(basename, "").count(".") == 1:
                        ret.append(fullname)
            return ret

    @staticmethod
    def DetermineLoadDir(loaddir, searchdir):
        if loaddir:
            loadDir = loaddir
        else:
            if not searchdir or not os.path.isdir(searchdir):
                searchdir = os.getcwd()
            loadDir = os.path.join(searchdir, "obj/CodeGenerator/Python/")
            while not os.path.isdir(loadDir):
                lastsearchdir = searchdir
                searchdir = os.path.abspath(os.path.join(searchdir, os.pardir))
                # checking if joining with pardir returns the same dir is the easiest way to
                # determine if we're at root of filesystem, and our upward search has to stop.
                if lastsearchdir == searchdir:
                    # check for global path to generated code
                    if os.path.isdir("/opt/msgtools/obj/CodeGenerator/Python/"):
                        loadDir = "/opt/msgtools/obj/CodeGenerator/Python/"
                        break
                    else:
                        # if we're at root of filesystem, just give up!
                        loadDir = None
                        raise RuntimeError('''
ERROR! Auto-generated python code not found!
cd to a directory downstream from a parent of obj/CodeGenerator/Python
or specify that directory with --msgdir=PATH''')
                loadDir = searchdir + "/obj/CodeGenerator/Python/"
                if Messaging.debug:
                    print("search for objdir in " + loadDir)

        # Normalize our load directory path, and load message modules on the path and
        # fix the sys path for future imports
        loadDir = os.path.normpath(loadDir)
        
        # add these to the system path, to allow application code to import header and message files directly.
        sys.path.append(loadDir)
        sys.path.append(os.path.join(loadDir, "headers"))
        
        Messaging.objdir = os.path.abspath(os.path.join(loadDir,"../.."))

        return loadDir


    @staticmethod
    def LoadHeader(loaddir=None, searchdir=None, headerName="NetworkHeader"):
        loadDir = Messaging.DetermineLoadDir(loaddir, searchdir)

        # if we didn't find valid auto-generated code, this will cause an import error!
        # the fix is to point to valid auto-generated code!
        headerModule = importlib.import_module("headers." + headerName)

        # Set the global header class
        Messaging.hdr = getattr(headerModule, headerName)

    @staticmethod
    def LoadAllMessages(loaddir=None, searchdir=None, headerName="NetworkHeader"):
        '''
        This dynamically loads all generated message code, with each message registering itself
        with the Messaging class on import.  This allows us to support a number of utility
        functions for message creation, introspection, class lookup, etc

        loaddir - If set we use this as the base directory for generated message code.  Overrides searchdir.
        searchdir - If set we start a search for 'obj/CodeGenerator/Python' from searchdir, and moving up 
            to each parent directory until we find this folder.  If loaddir and searchdir are not set
            we default to the current directory.
        debug - print debugging information to the console.
        headerName - the default header to use when processing messages.  We assume this module resides
            in a 'headers' folder below our base message root directory.
        '''
        if Messaging.hdr != None:
            print("Messaging.LoadAllMessages() called more than once.")
            print("Returning immediately instead of re-executing it!")
            return

        loadDir = Messaging.DetermineLoadDir(loaddir, searchdir)
        
        # if we didn't find valid auto-generated code, this will cause an import error!
        # the fix is to point to valid auto-generated code!
        headerModule = importlib.import_module("headers." + headerName)

        # Set the global header class
        Messaging.hdr = getattr(headerModule, headerName)

        # specify our header size, to come from the generated header we imported
        Messaging.hdrSize = Messaging.hdr.SIZE
        
        cache_filename = os.path.join(loadDir, 'msglib_cache.json')
        try:
            with open(cache_filename, 'r') as fp:
                msglibinfo = json.load(fp)
                Messaging.MsgNameFromID     = msglibinfo["MsgNameFromID"]
                Messaging.MsgIDFromName     = msglibinfo["MsgIDFromName"]
                Messaging.MsgModuleFromName = msglibinfo["MsgModuleFromName"]
                for name in Messaging.MsgIDFromName:
                    # Initialize all class lookups to None, unless they are already set.
                    # This is a signal for MessageNameLoader to do a lookup, while
                    # still providing keys() for users to access.
                    if not name in Messaging.MsgClassFromName:
                        Messaging.MsgClassFromName[name] = None
                    
                    # the below is used to populate the hierarchy of dot-notation,
                    # *except* for the final class at the end (which will be done
                    # with lazy lookup)
                    
                    # split up the name between periods, and add it to the Messaging object so it
                    # can be accessed via dot notation
                    nameParts = name.split(".")
                    messagingVars = vars(Messaging.Messages)
                    basename = ""
                    for namePart in nameParts[:-1]:
                        if basename:
                            basename = basename + "." + namePart
                        else:
                            basename = namePart
                        if namePart and not namePart in messagingVars:
                            messagingVars[namePart] = MessageAttributeLoader(basename)
                        messagingVars = vars(messagingVars[namePart])
                
                # now that cache is built, check if we need to load/reload
                # any files because their python is newer than the cache
                cache_file_timestamp = os.path.getmtime(cache_filename)
        except FileNotFoundError:
            cache_file_timestamp = 0

        write_cache_file = Messaging.LoadDir(loadDir, loadDir, cache_file_timestamp=cache_file_timestamp)

        if write_cache_file:
            # store a cache message ID, name, and file data
            msglibinfo = {
                "MsgNameFromID":     Messaging.MsgNameFromID,
                "MsgIDFromName":     Messaging.MsgIDFromName,
                "MsgModuleFromName": Messaging.MsgModuleFromName
            }
            with open(cache_filename, 'w') as fp:
                json.dump(msglibinfo, fp)

    @staticmethod
    def LoadDir(loadDir, rootDir, cache_file_timestamp=0):
        loaded = False
        for filename in os.listdir(loadDir):
            filepath = loadDir + '/' + filename
            if os.path.isdir(filepath):
                if filename != 'headers':
                    if Messaging.debug:
                        print("descending into directory ", filepath)
                    loaded_subdir = Messaging.LoadDir(filepath, rootDir, cache_file_timestamp)
                    loaded = loaded or loaded_subdir
            elif filename.endswith('.py'):
                file_timestamp = os.path.getmtime(filepath)
                if file_timestamp > cache_file_timestamp:
                    if Messaging.debug and cache_file_timestamp != 0:
                        print("%s out of date (%f > %f), reloading" % (filename, file_timestamp, cache_file_timestamp))
                    # Build an import friendly module name
                    # Could've just passed the root length to be more efficient
                    # but felt the easier readability was better than the miniscule
                    # performance gain we could've had
                    modulename = filepath[len(rootDir)+1:]
                    modulename = modulename[0:modulename.rfind('.py')]
                    modulename = modulename.replace('/', '.')
                    if Messaging.debug:
                        print("loading module "+modulename)
        
                    loaded = True
                    importlib.import_module(modulename)
        return loaded

    @staticmethod
    def Register(name, id, classDef):
        'add message to the global hash table of names by ID, and IDs by name'

        hexid = hex(id)
        
        if Messaging.debug:
            print("Registering", name, "as",hexid)
        
        if(hexid in Messaging.MsgNameFromID and Messaging.MsgNameFromID[hexid] != name):
            print("WARNING! Trying to define message ", name, " for ID ", hexid, ", but ", Messaging.MsgNameFromID[hexid], " already uses that ID")
        
        Messaging.MsgNameFromID[hexid] = name

        Messaging.AddAlias(name, id, classDef)

    @staticmethod
    def AddAlias(name, id, classDef):
        'add message to the global hash table of IDs by name'
        
        hexid = hex(id)

        if(name in Messaging.MsgIDFromName and Messaging.MsgIDFromName[name] != hexid):
            print("WARNING! Trying to define message %s for ID %s(%d), but %s(%d) already uses that name" % (name, hexid, id, Messaging.MsgIDFromName[name], int(Messaging.MsgIDFromName[name], 0)))

        Messaging.MsgIDFromName[name] = hexid
        Messaging.MsgClassFromName[name] = classDef
        Messaging.MsgModuleFromName[name] = classDef.__module__

        # split up the name between periods, and add it to the Messaging object so it
        # can be accessed via dot notation
        nameParts = name.split(".")
        messagingVars = vars(Messaging.Messages)
        basename = ""
        for namePart in nameParts[:-1]:
            if basename:
                basename = basename + "." + namePart
            else:
                basename = namePart
            if namePart and not namePart in messagingVars:
                messagingVars[namePart] = MessageAttributeLoader(basename)
            messagingVars = vars(messagingVars[namePart])
        messagingVars[nameParts[-1]] = classDef

    @staticmethod
    def MsgFactory(msg, name=None):
        '''Create a new message instance from the given msg

            msg - the header and message payload (as a NetworkHeader)

            name - will override msg - create an empty message by name

            return an instance of the message class for this msg, or UnknownMsg
                if we can't find this message type
        '''
        # TODO: We could do a type check here and support a NetworkHeader or
        # simple raw buffer
        if name is None:
            msgClass = Messaging.MsgClass(msg)
            return msgClass(msg.rawBuffer())

        if name not in Messaging.MsgClassFromName:
            return None

        return Messaging.MsgClassFromName[name]()

    @staticmethod
    def MsgClass(hdr):
        '''Find the message class for this message header

        hdr - the header for this message.  Need not be the full
            message, i.e. header + payload.

        return UnknownMsg if we can't find this message, otherwise
            a class reference to the message for the given header (by ID)
        '''
        msgId = hex(hdr.GetMessageID())

        if not msgId in Messaging.MsgNameFromID:
            #print("WARNING! No definition for ", msgId, "!\n")
            from msgtools.lib.unknownmsg import UnknownMsg
            msgClass = UnknownMsg
        else:
            msgName = Messaging.MsgNameFromID[msgId]
            msgClass = Messaging.MsgClassFromName[msgName]

        return msgClass

    @staticmethod
    def set(msg, fieldInfo, value, index=0):
        if("int" in fieldInfo.type or "enumeration" == fieldInfo.type):
            if isinstance(value, str):
                value = value.strip()
                if value.startswith("0x"):
                    value = int(value, 0)
        if("int" in fieldInfo.type):
            value = int(float(value))
        elif("float" in fieldInfo.type):
            value = float(value)
        
        if fieldInfo.count == 1:
            fieldInfo.set(msg, value)
        else:
            fieldInfo.set(msg, value, index)

    @staticmethod
    def get(msg, fieldInfo, index=0):
        try:
            if fieldInfo.count == 1:
                value = fieldInfo.get(msg)
            else:
                value = fieldInfo.get(msg, index)
            if "hex" == fieldInfo.units.lower():
                digits = fieldInfo.size*2
                value = ("0x%0"+str(digits)+"X") % (value)
        except struct.error:
            value = "UNALLOCATED"
        return value

    @staticmethod
    def getFloat(msg, fieldInfo, index=0):
        value = Messaging.get(msg, fieldInfo, index)
        if len(fieldInfo.enum) != 0:
            value = fieldInfo.enum[0].get(value, value)
        return float(value)

    @staticmethod
    def getAlert(msg, fieldInfo, index=0):
        value = Messaging.get(msg, fieldInfo, index)
        alert = 0
        try:
            floatVal = float(value)
            minVal = float(fieldInfo.minVal)
            maxVal = float(fieldInfo.maxVal)
            if(floatVal < minVal or floatVal > maxVal):
                alert = 1
        except ValueError:
            pass
        return alert

    @staticmethod
    def findFieldInfo(fieldInfos, name):
        for fi in fieldInfos:
            if name == fi.name:
                return fi
            for bfi in fi.bitfieldInfo:
                if name == bfi.name:
                    return bfi
        return None
    
    # Create a fake field.  This is useful if we want to compute data based on an
    # existing message's fields that we want to plot or log.
    @staticmethod
    def createFakeField(msg_class, name, units="", minVal="", maxVal="", description="Fake field!", count=1):
        def getFn(self, index=0):
            if count == 1:
                return self.fake_fields[name]
            else:
                return self.fake_fields[name][index]
        getFn.units = units
        getFn.default = ''
        getFn.minVal = minVal
        getFn.maxVal = maxVal
        getFn.offset = 0
        getFn.size = 0
        getFn.count = count
        def setFn(self, value, index=0):
            if count == 1:
                self.fake_fields[name] = value
            else:
                if not name in self.fake_fields:
                    self.fake_fields[name] = []
                self.fake_fields[name][index] = value
        setFn.units = units
        setFn.default = ''
        setFn.minVal = minVal
        setFn.maxVal = maxVal
        setFn.offset = 0
        setFn.size = 0
        setFn.count = count
        new_fi = FieldInfo(name, type="", units=units, minVal=minVal, maxVal=maxVal, description=description, get=getFn, set=setFn, count=count, bitfieldInfo=[], enum=[])
        msg_class.fields.append(new_fi)

    # This is composed of all the header fields that are not length, time, and any ID fields.
    # In some systems where one PC talks to one device, there may not be *any* fields that
    # contribute to Route.  In other systems it could be a device ID on a CANbus, a source
    # or destination field, or an IP Address or MAC address.
    # The reason MsgRoute exists is to distinguish between traffic with the same ID, that
    # is to, from, for, between different entities, so they aren't assumed to be two messages
    # in one time sequence.  This is necessary for showing lists of data (like msginspector),
    # showing plots of data as a time series, or showing the latest value of all data (msgscope).
    # As a performance optimization, the first time this function is called for a specific type
    # of header, we make a list of route fields, and then every time we need to generate a Route,
    # we loop through those fields.
    @staticmethod
    def HeaderRoute(hdr):
        try:
            routeFields = type(hdr).s_routeFields
        except AttributeError:
            routeFields = []
            for fieldInfo in hdr.fields:
                if fieldInfo.bitfieldInfo:
                    for bitfieldInfo in fieldInfo.bitfieldInfo:
                        if Messaging.IsRouteField(bitfieldInfo):
                            routeFields.append(bitfieldInfo)
                else:
                    if Messaging.IsRouteField(fieldInfo):
                        routeFields.append(fieldInfo)
            type(hdr).s_routeFields = routeFields
        # list comprehension to build the message route based on each of the route fields.
        msg_route = [str(route_field_info.get(hdr)) for route_field_info in routeFields]
        return msg_route

    @staticmethod
    def MsgRoute(msg):
        return Messaging.HeaderRoute(msg.hdr)
    
    # this sets up a message's header based on the given route
    #TODO Need special logic for headers with source and destination, because they need to be
    #TODO swapped if the message is coming or going.
    @staticmethod
    def SetMsgRoute(msg, msg_route):
        hdr = msg.hdr
        i = 0
        for fieldInfo in msg.hdr.fields:
            if fieldInfo.bitfieldInfo:
                for bitfieldInfo in fieldInfo.bitfieldInfo:
                    if Messaging.IsRouteField(bitfieldInfo):
                        Messaging.set(msg.hdr, bitfieldInfo, msg_route[i])
            else:
                if Messaging.IsRouteField(fieldInfo):
                    Messaging.set(msg.hdr, fieldInfo, msg_route[i])

    # fields count as Route fields unless they are part of the ID,
    # have "length" in their name, or match Time or Priority.
    @staticmethod
    def IsRouteField(fieldInfo):
        if fieldInfo.idbits != 0:
            return False
        if "length" in fieldInfo.name.lower():
            return False
        if fieldInfo.name == "Time":
            return False
        if fieldInfo.name == "Priority":
            return False
        return True
        
class BitFieldInfo(object):
    def __init__(self, name, type, units, minVal, maxVal, description, get, set, enum, idbits=0):
        self.name=name
        self.type=type
        self.units=units
        self.minVal=minVal
        self.maxVal=maxVal
        self.description=description
        self.get=get
        self.set=set
        self.count=1
        self.enum=enum
        self.idbits=idbits
        # add a couple fields from decorators of the 'get' function
        self.offset = int(get.offset)
        self.size = int(get.size)

    def exists(self, msg, index=0):
        return self.parent.exists(msg, index)

    def end_location(self, index=0):
        return self.parent.end_location(index)

class FieldInfo(object):
    def __init__(self, name, type, units, minVal, maxVal, description, get, set, count, bitfieldInfo, enum, idbits=0):
        self.name=name
        self.type=type
        self.units=units
        self.minVal=minVal
        self.maxVal=maxVal
        self.description=description
        self.get=get
        self.set=set
        self.count=count
        self.bitfieldInfo=bitfieldInfo
        self.enum=enum
        self.idbits=idbits
        # add a couple fields from decorators of the 'get' function
        self.offset = int(get.offset)
        self.size = int(get.size)
        # give our bitfields a reference to us
        for bfi in self.bitfieldInfo:
            bfi.parent = self
            # Set the count of the bitfield to our own count, so that bitfields
            # in arrays work properly.
            bfi.count = count

    def exists(self, msg, index=0):
        end_of_field = self.offset + self.size * (index+1)
        if msg.hdr.GetDataLength() >= end_of_field:
            return True
        return False
    
    def end_location(self, index=0):
        return self.offset + self.size * (index+1)



#!/usr/bin/env python3
import sys
#import argparse

class RowInfo:
    def __init__(self, arg):
        row_info = arg.split('=')
        self.row_name = row_info[0]
        self.col_search_strings = row_info[1].split(',')
        self.col_numbers = []
        self.col_number_offsets = {}
        for i in range(len(self.col_search_strings)):
            name = self.col_search_strings[i]
            if '+' in name:
                parts = self.col_search_strings[i].split("+")
                self.col_search_strings[i] = parts[0]
                self.col_number_offsets[self.col_search_strings[i]] = int(parts[1])
                
        self.got_header = False
        self.column_names = []
    
    def ColumnCount(self):
        return len(self.col_search_strings)
    
    def GotHeader(self):
        return self.got_header

    def ProcessData(self, line, blank_columns=0):
        if self.row_name in line:
            cols = line.split(',')
            if self.row_name in cols[1]:
                if self.got_header:
                    output = [cols[0], cols[1]] + [''] * blank_columns
                    # normal row
                    for col_number in self.col_numbers:
                        output.append(cols[col_number])
                    return ','.join(output)
                else:
                    # header row
                    self.got_header = True
                    for col_name in self.col_search_strings:
                        col_number = 0
                        for col in cols:
                            if col_name in col:
                                if not col_number in self.col_numbers:
                                    self.column_names.append(col)
                                    value = col_number
                                    if col_name in self.col_number_offsets:
                                        value += self.col_number_offsets[col_name]
                                    self.col_numbers.append(value)
                            col_number+=1
        return None
        

def GotAllHeaders(row_infos):
    ret = True
    for r in row_infos:
        if not r.GotHeader():
            ret = False
    return ret

def PrintHeader(row_infos):
    ret = ["Time","MessageName"]
    for r in row_infos:
        ret += r.column_names
    return ','.join(ret)
    
def main(args=None):
    #parser = argparse.ArgumentParser(description="Tool to extract rows and columns from a CSV")
    if len(sys.argv) < 3:
        print("Like this:")
        print(sys.argv[0] + ' MsgName1=Field1,Field[+1] MsgName2=FieldA,FieldB[+2]')
        sys.exit(-1)
    file_name = sys.argv[1]
    row_infos = []
    for arg in sys.argv[2:]:
        row_infos.append(RowInfo(arg))
    
    got_all_headers = False
    output = ""
    with open(file_name, 'r') as csv_file:
        lines = csv_file.read().splitlines()
        for line in lines:
            blank_columns = 0
            for r in row_infos:
                ret = r.ProcessData(line, blank_columns)
                if ret:
                    if output:
                        output = output + '\n' + ret
                    else:
                        output = ret
                blank_columns += r.ColumnCount()
            if got_all_headers:
                if output:
                    print(output)
                    output = ""
            else:
                if GotAllHeaders(row_infos):
                    got_all_headers = True
                    print(PrintHeader(row_infos))
                    print(output)
                    output = ""
                else:
                    pass
                        
            
    

# main starts here
if __name__ == '__main__':
    main()

#grep MotorStatus 20200117-093703_deploy.csv | cut -d, -f 1,29,41


import argparse
import os
import socket
import sys

if sys.version_info<(3,4):
    raise SystemExit('''\n\nSorry, this code need Python 3.4 or higher.\n
To avoid incorrectly using python2 in your path, you may want to try launching by:
    ./path/to/script/ScriptName.py
NOT
    python path/to/script/ScriptName.py\n''')

from PyQt5 import QtGui, QtWidgets, QtCore, QtNetwork
from PyQt5.QtWidgets import QInputDialog, QLineEdit

from .messaging import Messaging, TimestampFixer

class App(QtWidgets.QMainWindow):
    RxMsg = QtCore.pyqtSignal(object)
    statusUpdate = QtCore.pyqtSignal(str)
    connectionChanged = QtCore.pyqtSignal(bool)

    @classmethod
    def addBaseArguments(cls, parser):
        '''
        Adds base app arguments to the provided ArgParser
        skipFiles - if True we won't provide an agument for a files list
        returns the parser
        '''
        parser.add_argument('--connectionType', choices=['socket', 'qtsocket', 'file'], default='qtsocket',
            help='Specify the type of connection we are establishing as a message pipe/source.')
        parser.add_argument('--connectionName', default='127.0.0.1:5678',
            help='''The connection name.  For socket connections this is an IP and port e.g. 127.0.0.1:5678.  
                    You may prepend ws:// to indicate you want to use a Websocket instead.  For file 
                    connection types, this is the filename to use as a message source.
                    This parameter is overridden by the --ip and --port options.''')
        parser.add_argument('--ip', help='The IP address for a socket connection.   Overrides connectionName.')
        parser.add_argument('--port', type=int, help='The port for a socket connection. Overrides connectionName.')
        parser.add_argument('--msg', nargs='+', default=set(), help='''A space delimited list of white list messages to process.
                    All messages outside of this list will be ignored.  For example: --msg TestCase1 Network.Note 
                    Taxonomy/Candidae.AFox''')
        parser.add_argument('--msgdir', help=''''The directory to load Python message source from.''')
        parser.add_argument('--serial', action='store_true', help='Set if you want to use a SerialHeader instead of a NetworkHeader.')
        parser.add_argument('--log', help='The log file type (csv/json/bin) or complete log file name.')

        return parser
    
    def __init__(self, name, args):
        '''Initializer - this is used more as a decorator in the MsgTools suite than an actual
        standalone class.

        name - the name of the application - usually displayed in the UI window header
        args - argparse.Namespace of command line options.  See getArgParser for what we provide
               by default as part of this base app framework.  A command line tool may act as a 
               parent ArgumentParser and provide additional options.
        '''

        # If the caller skips adding base arguments we need to patch up args
        args.serial = None if hasattr(args, 'serial') == False else args.serial
        args.msg = None if hasattr(args, 'msg') == False else args.msg
        args.msgdir = None if hasattr(args, 'msgdir') == False else args.msgdir

        # default to Network, unless we have a input filename that contains .txt
        headerName = "NetworkHeader"
        if args.serial or (args.connectionType=='file' and os.path.splitext(args.connectionType)[1].lower() == '.txt'):
            headerName = "SerialHeader"

        self.name = name
        
        # persistent settings
        self.settings = QtCore.QSettings("MsgTools", name)
        
        # rx buffer, to receive a message with multiple signals
        self.rxBuf = bytearray()
        
        self.allowedMessages = set(args.msg)
        
        # flag that indicates if we're connected
        self.connected = False
        
        # directory to load messages from.
        msgLoadDir = None

        # connection modes
        ip = ""
        port = ""

        if args.connectionType is not None:
            self.connectionType = args.connectionType
        if args.connectionName is not None:
            self.connectionName = args.connectionName
        if args.ip is not None:
            ip = args.ip
        if args.msgdir:
            msgLoadDir = args.msgdir
            self.msgdir = args.msgdir
        
        # if either --ip or --port were used, override connectionName
        if args.ip is not None or args.port is not None:
            self.connectionName = str(ip)+":"+str(port)
        
        # initialize the read function to None, so it's not accidentally called
        self.readBytesFn = None

        try:
            Messaging.LoadAllMessages(searchdir=msgLoadDir, headerName=headerName)
        except RuntimeError as e:
            print(e)
            quit()
        except:
            import traceback
            print(traceback.format_exc())
            quit()

        if args.msg is not None:
            # Validate all message names are valid
            for msg in args.msg:
                if msg not in Messaging.MsgIDFromName:
                    print('{0} is not a valid message name!'.format(msg))
                    sys.exit(1)

        # object to fix timestamps of messages with no timestamp set
        self.timestamp_fixer = TimestampFixer()

        self.logFileType = None
        self.logFile = None
        if args.log is not None:
            self.startLog(args.log)

        port = args.port
        
        self.OpenConnection()

    def CloseConnection(self):
        if hasattr(self, 'connection') and (self.connectionType.lower() == "socket" or self.connectionType.lower() == "qtsocket"):
            if "ws:" in self.connectionName:
                if self.connection:
                    self.connection.close()
                    self.connection = None
            else:
                if self.connection:
                    self.connection.disconnectFromHost()
                    self.connection = None

    # this function opens a connection, and returns the connection object.
    def OpenConnection(self):
        self.CloseConnection()

        if(self.connectionType.lower() == "socket" or self.connectionType.lower() == "qtsocket"):
            if "ws:" in self.connectionName:
                connectionName = self.connectionName.replace("ws://","")
                (ip, port) = connectionName.rsplit(":",1)
                if ip == None or ip == "":
                    ip = "127.0.0.1"

                if port == None or port == "":
                    port = "5679"
                connectionName = "ws://"+ip+":"+port
                from PyQt5.QtWebSockets import QWebSocket
                from PyQt5.QtCore import QUrl
                self.connection = QWebSocket()
                print("opening websocket " + connectionName)
                self.connection.open(QUrl(connectionName))
                self.connection.binaryMessageReceived.connect(self.processBinaryMessage)
                self.sendBytesFn = self.connection.sendBinaryMessage
            else:
                #print("opening TCP socket " + self.connectionName)
                (ip, port) = self.connectionName.rsplit(":",1)
                if ip == None or ip == "":
                    ip = "127.0.0.1"

                if port == None or port == "":
                    port = "5678"
                
                port = int(port)

                #print("ip is ", ip, ", port is ", port)
                self.connection = QtNetwork.QTcpSocket(self)
                self.connection.error.connect(self.displayConnectError)
                ret = self.connection.readyRead.connect(self.readRxBuffer)
                self.connection.connectToHost(ip, port)
                self.readBytesFn = self.connection.read
                self.sendBytesFn = self.connection.write
                #print("making connection returned", ret, "for socket", self.connection)
            self.connection.connected.connect(self.onConnected)
            self.connection.disconnected.connect(self.onDisconnect)
        elif(self.connectionType.lower() == "file"):
            try:
                self.connection = open(self.connectionName, 'rb')
                self.readBytesFn = self.connection.read
                self.sendBytesFn = self.connection.write
            except IOError:
                print("\nERROR!\ncan't open file ", self.connectionName)
                sys.exit(1)
        else:
            print("\nERROR!\nneed to specify socket or file")
            sys.exit()

        self.connection

    def startLog(self, log_name, log_suffix=''):
        # hash table of booleans for if each type of message has had it's header
        # put into this log file already.
        self.loggedMsgHeader = {}
        if log_name.endswith('csv'):
            self.logFileType = "csv"
        elif log_name.endswith('json'):
            self.logFileType = "json"
        elif log_name.endswith('bin') or log_name.endswith('log'):
            self.logFileType = "bin"
        else:
            print("ERROR!  Invalid log type " + log_name)
        if "." in log_name:
            # if there's a ., assume they specified an exact filename to use
            logFileName = log_name
        else:
            # if not, generate a filename based on current date/time
            currentDateTime = QtCore.QDateTime.currentDateTime()
            logFileName = currentDateTime.toString("yyyyMMdd-hhmmss") + log_suffix + "." + self.logFileType
        self.logFile = QtCore.QFile(logFileName)
        self.logFile.open(QtCore.QIODevice.Append)

    def stopLog(self):
        if self.logFile is not None:
            self.logFile.close()
            self.logFile = None

    def onConnected(self):
        self.connected = True
        self.connectionChanged.emit(True)
        # send a connect message
        connectMsg = Messaging.Messages.Network.Connect()
        connectMsg.SetName(self.name)
        self.SendMsg(connectMsg)
        # if the app has it's own function to happen after connection, assume it will set subscriptions to what it wants.
        try:
            fn = self.onAppConnected
        except AttributeError:
            # send a subscription message
            subscribeMsg = Messaging.Messages.Network.MaskedSubscription()
            self.SendMsg(subscribeMsg)
            #self.statusUpdate.emit('Connected')
        else:
            self.onAppConnected()
    
    def onDisconnect(self):
        self.connected = False
        self.connectionChanged.emit(False)
        #self.statusUpdate.emit('NOT Connected')
    
    def displayConnectError(self, socketError):
        self.connected = False
        self.connectionChanged.emit(False)
        self.statusUpdate.emit('Not Connected('+str(socketError)+'), '+self.connection.errorString())

    # Qt signal/slot based reading of websocket
    def processBinaryMessage(self, bytes):
        if isinstance(bytes, bytearray):
            hdr = Messaging.hdr(bytes)
        else:
            # Assume this is a QByteArray from a websocket
            hdr = Messaging.hdr(bytes.data())

        # if we got this far, we have a whole message! Emit the signal
        # if the message is whitelisted
        msg = Messaging.MsgFactory(hdr)
        if self._messageAllowed(msg):
            self.RxMsg.emit(msg)
            self.logMsg(msg)

    def _messageAllowed(self, msg):
        '''Check msg against the list of allowed messages.

        return True if all messages are allowed, or the message is
        in the message white list.'''
        retVal = True
        if len(self.allowedMessages) > 0:
            if not msg.MsgName() in self.allowedMessages:
                retVal = False

        return retVal

    # Qt signal/slot based reading of TCP socket
    def readRxBuffer(self):
        input_stream = QtCore.QDataStream(self.connection)
        while(self.connection.bytesAvailable() > 0):
            # read the header, unless we have the header
            if(len(self.rxBuf) < Messaging.hdrSize):
                #print("reading", Messaging.hdrSize - len(self.rxBuf), "bytes for header")
                self.rxBuf += input_stream.readRawData(Messaging.hdrSize - len(self.rxBuf))
                #print("have", len(self.rxBuf), "bytes")
            
            # if we still don't have the header, break
            if(len(self.rxBuf) < Messaging.hdrSize):
                print("don't have full header, quitting")
                return
            
            hdr = Messaging.hdr(self.rxBuf)
            
            # need to decode body len to read the body
            bodyLen = hdr.GetDataLength()
            
            # read the body, unless we have the body
            if(len(self.rxBuf) < Messaging.hdrSize + bodyLen):
                #print("reading", Messaging.hdrSize + bodyLen - len(self.rxBuf), "bytes for body")
                self.rxBuf += input_stream.readRawData(Messaging.hdrSize + bodyLen - len(self.rxBuf))
            
            # if we still don't have the body, break
            if(len(self.rxBuf) < Messaging.hdrSize + bodyLen):
                print("don't have full body, quitting")
                return

            # Process what we assume to be a full message...            
            self.processBinaryMessage(self.rxBuf)

            # then clear the buffer, so we start over on the next message
            self.rxBuf = bytearray()
    
    def SendMsg(self, msg):
        bufferSize = len(msg.rawBuffer().raw)
        computedSize = Messaging.hdrSize + msg.hdr.GetDataLength()
        if(computedSize > bufferSize):
            msg.hdr.SetDataLength(bufferSize - Messaging.hdrSize)
            print("app.py: Truncating message %s from %d to %d bytes" % (msg.__class__.__name__, computedSize, bufferSize))
        if(computedSize < bufferSize):
            # don't send the *whole* message, just a section of it up to the specified length
            self.sendBytesFn(msg.rawBuffer().raw[0:computedSize])
        else:
            self.sendBytesFn(msg.rawBuffer().raw)
        # Log all sent messages
        self.logMsg(msg)

    def logMsg(self, msg):
        if self.logFile:
            original_timestamp = self.timestamp_fixer.fix_timestamp(msg.hdr)
            log = ''
            if self.logFileType == "csv":
                if not msg.MsgName() in self.loggedMsgHeader:
                    self.loggedMsgHeader[msg.MsgName()] = True
                    log = msg.csvHeader(timeColumn=True)+'\n'
                log += msg.toCsv(timeColumn=True)+'\n'
                log = log.encode('utf-8')
            elif self.logFileType == "json":
                if not msg.MsgName() in self.loggedMsgHeader:
                    self.loggedMsgHeader[msg.MsgName()] = True
                    log = msg.jsonHeader()
                log += msg.toJson(includeHeader=True)+'\n'
                log = log.encode('utf-8')
            elif self.logFileType == "bin":
                log = msg.rawBuffer().raw
            self.logFile.write(log)
            self.logFile.flush()
            self.timestamp_fixer.restore_timestamp(msg.hdr, original_timestamp)

    # this function reads messages (perhaps from a file, like in LumberJack), and calls the message handler.
    # unclear if it ever makes sense to use this in a application that talks to a socket or UART, because
    # it exits when there's no more data.  Perhaps it does in a CLI that's very procedural, like an automated
    # system test script?
    #
    # Note: This loop won't work with Qt socket signals.  For signals to work you have to exeucute the
    # application event loop.  To make this method work with QtSockets you would need to layer in 
    # the various waitFor* methods on the connection within this loop.
    def MessageLoop(self):
        msgCount=0
        startSeqField = Messaging.findFieldInfo(Messaging.hdr.fields, "StartSequence")
        if startSeqField == None:
            print("header contains no StartSequence")
        else:
            startSequence = int(Messaging.hdr.GetStartSequence.default)
            print("header contains StartSequence " + hex(startSequence))
        try:
            while (1):
                msgCount+=1
                self.rxBuf = self.readBytesFn(Messaging.hdrSize)
                
                if(len(self.rxBuf) != Messaging.hdrSize):
                    raise StopIteration
                
                hdr = Messaging.hdr(self.rxBuf)
                try:
                    start = hdr.GetStartSequence()
                    if(start != startSequence):
                        print("Error on message " + str(msgCount) + ". Start sequence invalid: 0x" + format(start, '02X'))
                        # resync on start sequence
                        bytesThrownAway = 0
                        while (1):
                            self.rxBuf += self.readBytesFn(1)
                            self.rxBuf = self.rxBuf[1:]
                            if(len(self.rxBuf) != Messaging.hdrSize):
                                raise StopIteration
                            bytesThrownAway += 1
                            start = hdr.GetStartSequence()
                            if(start == startSequence):
                                print("Resynced after " + str(bytesThrownAway) + " bytes")
                                break
                    headerChecksum = hdr.GetHeaderChecksum()
                    bodyChecksum = hdr.GetBodyChecksum(self)
                except AttributeError:
                    pass

                # need to decode body len to read the body
                bodyLen = hdr.GetDataLength()
                
                # read the body
                self.rxBuf += self.readBytesFn(bodyLen)
                if(len(self.rxBuf) != Messaging.hdrSize + bodyLen): break
                
                # create a new header object with the appended body
                hdr = Messaging.hdr(self.rxBuf)

                # got a complete message, call the callback to process it if
                # the message is on our white list
                msg = Messaging.MsgFactory(hdr)
                if self._messageAllowed(msg):
                    self.ProcessMessage(msg)
        except StopIteration:
            print("found end of file, exited")




import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

oneOutputFilePerMsg = True

def endian_string():
    if MsgParser.big_endian:
        return ''
    else:
        return ', true'

def paramType(field):
    fieldTypeDict = \
    {"uint64":"Ulong", "uint32":"UInt","uint16": "UShort",   "uint8": "UByte",
     "int64":"Long",   "int32":"Int",  "int16": "Short",  "int8": "Byte",
      "float64":"Double", "float32":"Float"}
    typeStr = field["Type"]
    return fieldTypeDict[typeStr]

def fieldType(field):
    fieldTypeDict = \
    {"uint64":"Ulong", "uint32":"UInt","uint16": "UShort",   "uint8": "UByte",
     "int64":"Long",   "int32":"Int",  "int16": "Short",  "int8": "Byte",
      "float64":"Double", "float32":"Float"}
    typeStr = field["Type"]
    return fieldTypeDict[typeStr]

def bitParamType(field, bitfield):
    ret = paramType(field)
    if "Offset" in bitfield or "Scale" in bitfield:
        return typeForScaledInt(field)
    return ret

def isUnsigned(field):
    typeStr = field["Type"]
    if typeStr == "uint8" or typeStr == "uint16" or typeStr == "uint32":
        return True
    return False

def literalSuffix(field):
    typeStr = field["Type"]
    if  "Offset" in field or "Scale" in field:
        if typeForScaledInt(field) == "Float":
            return "f"
        return ""
    if typeStr == "float32":
        return "f"
    if typeStr == "uint8" or typeStr == "uint16" or typeStr == "uint32":
        return "u"
    return ""

def literalSuffixForBitfield(field, bits):
    typeStr = field["Type"]
    if  "Offset" in field or "Scale" in bits:
        if typeForScaledInt(field) == "Float":
            return "f"
    return literalSuffix(field)


def fnHdr(field):
    ret = "// %s %s, (%s to %s)" % (MsgParser.fieldDescription(field), MsgParser.fieldUnits(field), MsgParser.fieldMin(field), MsgParser.fieldMax(field))
    return ret

def typeForScaledInt(field):
    numBits = fieldNumBits(field)
    if numBits > 24:
        return "Double"
    return "Float"

# for floats, append tag such as 'f' to constants to eliminate compiler warnings
def fieldScale(field):
    if "Scale" in field:
        ret = field["Scale"]
        if typeForScaledInt(field) == "Float":
            ret = str(ret) + "f"
    return ret

def fieldOffset(field):
    if "Offset" in field:
        ret = field["Offset"]
        if typeForScaledInt(field) == "Float":
            ret = str(ret) + "f"
    return ret

def getMath(x, field, cast, floatTag=""):
    ret = x
    if cast and ("Offset" in field or "Scale" in field):
        ret = "%s.toInt().to%s()" % (ret, cast)
    if "Scale" in field:
        ret = "(%s * %s)" % (ret, fieldScale(field))
    if "Offset" in field:
        ret = "(%s + %s)" % (ret, fieldOffset(field))
    if typeForScaledInt(field) == "Float":
        ret = ret+".toFloat()"
    return ret

def setMath(x, field, cast=None, floatTag=""):
    ret = x
    if "Offset" in field:
        ret = "(%s - %s)" % (ret, fieldOffset(field))
    if "Scale" in field:
        ret = "(%s / %s)" % (ret, fieldScale(field))
    if cast and ("Offset" in field or "Scale" in field):
        ret = "%s.to%s()" % (ret, cast)
    return ret

def getFn(field):
    loc = str(MsgParser.fieldLocation(field))
    param = ""
    if MsgParser.fieldCount(field) > 1:
        loc += " + index*" + str(MsgParser.fieldArrayElementOffset(field))
        param += "index: Int"
    retType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        retType = typeForScaledInt(field)
    ret = '''\
%s
fun get%s(%s): %s {
''' % (fnHdr(field), field["Name"], param, retType)
    access = "data.get%s(%s%s)" % (fieldType(field), loc, endian_string())
    if ("Offset" in field or "Scale" in field):
        ret += '    val valI : '+fieldType(field)+' = '+access+'\n'
        access = getMath("valI", field, "Double")
        ret += '    val valD = '+access+'\n'
        ret += '    return valD\n'
    else:
        ret += '    return %s\n' % (access)
    ret += '}'
    if MsgParser.fieldUnits(field) == "ASCII" and MsgParser.fieldCount(field) > 1 and (field["Type"] == "uint8" or field["Type"] == "int8"):
        ret += '''
%s
fun get%sString(): String {
    var value = ""
    for (i in 0 until minOf(%s, header.getDataLength().toInt() - %s)) {
        val nextChar = get%s(i)
        if (nextChar == 0.toUByte()) {
            break
        }
        value += nextChar.toByte().toChar()
    }
    return value
}''' % (fnHdr(field), field["Name"], str(MsgParser.fieldCount(field)), str(MsgParser.fieldLocation(field)), field["Name"])
    return ret

def setFn(field):
    paramType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        paramType = typeForScaledInt(field)
    valueString = setMath("value", field, fieldType(field))
    param = "value: " + paramType
    loc = str(MsgParser.fieldLocation(field))
    if MsgParser.fieldCount(field) > 1:
        loc += " + index*" + str(MsgParser.fieldArrayElementOffset(field))
        param += ", index: Int"
    ret = '''\
%s
fun set%s(%s) {
    data.put%s(%s, %s%s)
}''' % (fnHdr(field), field["Name"], param, fieldType(field), loc, valueString, endian_string())
    return ret

def getBitsFn(field, bits, bitOffset, numBits):
    # Kotlin currently only supports bitwise operators for Int, so we are gonna convert the inputs to Int first, 
    # apply the bit operators, and then convert the final result to the actual type
    access = "((get%s().toInt() ushr %s) and %s).to%s()" % (field["Name"], str(bitOffset), MsgParser.Mask(numBits), bitParamType(field, bits))
    param = ""
    ret = '''\
%s
fun get%s(%s): %s {
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), param, bitParamType(field, bits))
    if ("Offset" in bits or "Scale" in bits):
        ret += '    val valI = '+access+'\n'
        access = getMath("valI", bits, "Double")
        ret += '    val valD = '+access+'\n'
        ret += '    return valD\n'
    else:
        ret += '    return %s\n' % (access)

    ret += '}'
    return ret

def setBitsFn(field, bits, bitOffset, numBits):
    param = "value: " + bitParamType(field, bits)
    ret = '''\
%s
fun set%s(%s) {
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), param)
    valueString = setMath("value", bits)
    # Kotlin currently only supports bitwise operators for Int, so we are gonna convert the inputs to Int first, 
    # apply the bit operators, and then convert the final result to the actual type
    ret += '''\
    var valI = get%s().toInt() // read
    valI = valI and (%s shl %s).inv() // clear our bits
    valI = valI or ((%s.toInt() and %s) shl %s) // set our bits
    set%s(valI.to%s()) // write
''' % (field["Name"], MsgParser.Mask(numBits), str(bitOffset), valueString, MsgParser.Mask(numBits), str(bitOffset), field["Name"], paramType(field))
    ret += '}'
    return ret

def accessors(msg):
    gets = []
    sets = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            gets.append(getFn(field))
            sets.append(setFn(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    gets.append(getBitsFn(field, bits, bitOffset, numBits))
                    sets.append(setBitsFn(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return gets+sets

def initField(field):
    if "Default" in field:
        if MsgParser.fieldCount(field) > 1:
            ret = "for (i in 0 until "+str(MsgParser.fieldCount(field))+") {\n"
            ret += "    set" + field["Name"] + "(" + str(field["Default"]) + literalSuffix(field) + ", i)\n"
            ret += "}\n"
            return ret
        else:
            return  "set" + field["Name"] + "(" + str(field["Default"]) + literalSuffix(field) + ")"
    return ""

def initBitfield(field, bits):
    if "Default" in bits:
        return  "set" + MsgParser.BitfieldName(field, bits) + "(" +str(bits["Default"]) + literalSuffixForBitfield(field, bits) + ")"
    return ""

def initCode(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInit = initField(field)
            if fieldInit:
                ret.append(fieldInit)
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits = initBitfield(field, bits)
                    if bits:
                        ret.append(bits)

    return ret

def optionValue(option):
    ret = int(option["Value"])
    if ret > 2**31:
        return str(ret)+"L"
    return str(ret)

def enums(e):
    ret = ""
    for enum in e:
        ret +=  "enum class " + enum["Name"]+"(override val value: Long): MessageEnum {\n"
        ret += "    "
        for option in enum["Options"]:
            ret += OptionName(option)+"("+str(optionValue(option)) + '), '
        ret = ret[:-2]
        ret += ";\n"
        ret += '''\
    companion object {
        fun construct(value: Long): %s {
            return values().first { it.value == value }
        }
    }
}\n''' % (enum["Name"])
    return ret

def fieldMin(field):
    val = MsgParser.fieldMin(field)
    ret = str(val)
    if ret == 'DBL_MIN':
        ret = 'Double.MIN_VALUE'
    elif ret == 'FLT_MIN':
        ret = 'Float.MIN_VALUE'
    return ret

def fieldMax(field):
    val = MsgParser.fieldMax(field)
    ret = str(val)
    if ret == 'DBL_MAX':
        ret = 'Double.MAX_VALUE'
    elif ret == 'FLT_MAX':
        ret = 'Float.MAX_VALUE'
    return ret

def genericInfo(field, loc, type):
    params  = '    const val loc = ' + loc + '\n'
    params += '    val max = ' + str(fieldMax(field)) + '\n'
    params += '    val min = ' + str(fieldMin(field)) + '\n'
    params += '    const val units = "' + str(MsgParser.fieldUnits(field)) + '"' + '\n'
    params += '    const val count = ' + str(MsgParser.fieldCount(field)) + '\n'
    if "Default" in field:
        params += '    const val '+type+' defaultValue = ' + str(field["Default"]) + "\n" 
    if "Scale" in field:
        params += '    const val scale = ' + str(field["Scale"]) + '\n'
    if "Offset" in field:
        params += '    const val offset = ' + str(field["Offset"]) + '\n'
    return params
    
def fieldInfo(field):
    retType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        retType = typeForScaledInt(field)

    params  = 'object ' + field["Name"] + 'FieldInfo {\n'
    params += genericInfo(field, str(MsgParser.fieldLocation(field)), retType)
    params += '}\n'
    return params

def fieldBitsInfo(field, bits, bitOffset, numBits):
    retType = fieldType(field)
    if "Offset" in bits or "Scale" in bits:
        retType = typeForScaledInt(bits)

    params  = 'object ' + bits["Name"] + 'FieldInfo {\n'
    params += genericInfo(bits, str(MsgParser.fieldLocation(field)), retType)
    params += '    const val bitOffset = ' + str(bitOffset) + '\n'
    params += '    const val numBits   = ' + str(numBits) + '\n'
    params += '}\n'
    return params

def fieldInfos(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldInfo(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsInfo(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)
    
def declarations(msg, msg_enums):
    return []

def getMsgID(msg):
    prefix = "var ret: UInt = 0u\n"
    mods = ""
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "IDBits" in field:
                numBits = str(field["IDBits"])
                getStr = "get"+field["Name"]+"().toUInt()"
                if mods:
                    mods += "ret = ret shl " + numBits + "\n"
                mods += "ret += " + getStr + "\n"
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    if "IDBits" in bitfield:
                        numBits = str(bitfield["IDBits"])
                        getStr = "get"+BitfieldName(field, bitfield)+"().toUInt()"
                        if mods:
                            mods += "ret = ret shl " + numBits + "\n"
                        mods += "ret += " + getStr + "\n"
    return prefix + mods + "return ret"
    
def setMsgID(msg):
    ret = ""
    numBits = 0
    if "Fields" in msg:
        for field in reversed(msg["Fields"]):
            if "IDBits" in field:
                if numBits != 0:
                    ret += "\nret = ret shr " + str(numBits)+"\n"
                numBits = field["IDBits"]
                setStr = "ret and "+Mask(numBits)+literalSuffix(field)
                setStr = "("+setStr+").to%s()" % fieldType(field)
                ret +=  "set"+field["Name"]+"("+setStr+")"
            if "Bitfields" in field:
                for bitfield in reversed(field["Bitfields"]):
                    if "IDBits" in bitfield:
                        if numBits != 0:
                            ret += "\nid = ret shr " + str(numBits)+"\n"
                        numBits = bitfield["IDBits"]
                        setStr = "ret and "+Mask(numBits)+literalSuffix(field)
                        setStr = "("+setStr+").to%s()" % fieldType(field)
                        ret +=  "set"+bitfield["Name"]+"("+setStr+")"
    if ret.count('\n') > 1:
        ret = "var ret = id\n" + ret
    else:
        ret = "val ret = id\n" + ret
    return ret



#!/cygdrive/c/Python34/python.exe
import sys

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from msgtools.lib.messaging import Messaging

class NoEditDelegate(QStyledItemDelegate):
    def __init__(self, parent, item_for_delegate):
        QStyledItemDelegate.__init__(self, parent)
        self.item_for_delegate = item_for_delegate

    def createEditor(self, parent, option, index):
        item = self.item_for_delegate.itemFromIndex(index)
        if not item.editableColumn(index.column()):
            return None
        return QStyledItemDelegate.createEditor(self, parent, option, index)

class FieldItem(QTreeWidgetItem):
    def __init__(self, editable, tree_widget, msg, msg_key, fieldInfo, column_strings, index):
        if len(column_strings) == 0:
            column_strings = [None, fieldInfo.name, "", fieldInfo.units, fieldInfo.description]
            if index != None:
                column_strings[1] = "    [" + str(index) + "]"
        QTreeWidgetItem.__init__(self, None, column_strings)
        
        self.editable = editable
        self.fieldInfo = fieldInfo
        if index == None:
            self.fieldName = fieldInfo.name
        else:
            self.fieldName = "%s[%d]" % (fieldInfo.name, index)
        self.msg = msg
        self.msg_key = msg_key
        self.index = index

        if self.editable:
            self.setFlags(self.flags() | Qt.ItemIsEditable)
            
            if fieldInfo.type == "enumeration":
                self.overrideWidget = QComboBox()
                self.overrideWidget.addItems(fieldInfo.enum[0].keys())
                self.overrideWidget.activated.connect(self.overrideWidgetValueChanged)
                # there's some odd behavior in the UI when the box is editable :(
                # if you want it editable, uncomment this line, and play around and see if you like it
                #self.overrideWidget.setEditable(1)
                # store a hash table of value->ComboBoxIndex
                # this is NOT the same as value->enumIndex!
                self.comboBoxIndexOfEnum = {}
                for i in range(0, self.overrideWidget.count()):
                    self.comboBoxIndexOfEnum[self.overrideWidget.itemText(i)] = i

    def editableColumn(self, column):
        if not self.editable:
            return False
        if column != 2:
            return False
        return True

    def data(self, column, role):
        if column != 2:
            return super(FieldItem, self).data(column, role)

        alert = Messaging.getAlert(self.msg, self.fieldInfo, self.index)
        if role == Qt.FontRole:
            font = QFont()
            if alert == 1:
                font.setBold(1)
            return font
        if role == Qt.ForegroundRole:
            brush = QBrush()
            if alert == 1:
                brush.setColor(Qt.red)
            return brush

        if role == Qt.DisplayRole:
            value  = str(Messaging.get(self.msg, self.fieldInfo, self.index))
            
            try:
                self.overrideWidget
                valueAsString = str(Messaging.get(self.msg, self.fieldInfo, self.index))
                
                if valueAsString in self.comboBoxIndexOfEnum:
                    #self.overrideWidget.setCurrentText(valueAsString)
                    self.overrideWidget.setCurrentIndex(self.comboBoxIndexOfEnum[valueAsString])
                else:
                    #self.overrideWidget.setEditText(valueAsString)
                    self.overrideWidget.setCurrentIndex(-1)
            except AttributeError:
                pass
                
            return str(value)
            
        return super(FieldItem, self).data(column, role)

    def overrideWidgetValueChanged(self, value):
        valueAsString = self.overrideWidget.itemText(value)
        # set the value in the message/header buffer
        Messaging.set(self.msg, self.fieldInfo, valueAsString)

        # no need to reset UI to value read from message, if user picked value from drop down.
        # \todo: need to if they type something, though.
        
    def setData(self, column, role, value):
        if not self.editable:
            return

        if not column == 2:
            return

        if self.fieldInfo.name == "ID":
            return
        
        if self.fieldInfo.type == "int" and value.startswith("0x"):
            value = str(int(value, 0))

        # if user deletes the value, for anything besides a string,
        # return without setting the new value
        if self.fieldInfo.type != "string" and value == "":
            return

        # set the value in the message/header buffer
        Messaging.set(self.msg, self.fieldInfo, value, self.index)

        # get the value back from the message/header buffer and pass on to super-class' setData
        super(FieldItem, self).setData(column, role, str(Messaging.get(self.msg, self.fieldInfo, self.index)))

# This is used for the bits within the containing bit
class FieldBits(FieldItem):
    def __init__(self, editable, tree_widget, msg, msg_key, bitfieldInfo, index):
       column_strings = [None, "    " + bitfieldInfo.name, "", bitfieldInfo.units, bitfieldInfo.description]
       super(FieldBits, self).__init__(editable, tree_widget, msg, msg_key, bitfieldInfo, column_strings=column_strings, index=index)

class FieldBitfieldItem(FieldItem):
    def __init__(self, editable, tree_widget, msg, msg_key, fieldInfo, index):
        super(FieldBitfieldItem, self).__init__(editable, tree_widget, msg, msg_key, fieldInfo, column_strings=[], index=index)

        for bitfieldInfo in fieldInfo.bitfieldInfo:
            bitfieldBitsItem = FieldBits(self.editable, tree_widget, self.msg, msg_key, bitfieldInfo, index)
            self.addChild(bitfieldBitsItem)
            if self.editable:
                try:
                    bitfieldBitsItem.overrideWidget
                    tree_widget.setItemWidget(bitfieldBitsItem, 2, bitfieldBitsItem.overrideWidget)
                except AttributeError:
                    pass

# This is for holding a list of elements of an array, each of which is a
# regular FieldItem or a FieldBitfieldItem.
class FieldArrayItem(QTreeWidgetItem):
    CHILD_STRING_MAX_LENGTH = 12
    def __init__(self, editable, tree_widget, msg, msg_key, fieldInfo):
        column_strings = [None, fieldInfo.name, "", fieldInfo.units, fieldInfo.description]
        
        QTreeWidgetItem.__init__(self, None, column_strings)
        self.editable = editable

        self.fieldInfo = fieldInfo
        self.fieldName = fieldInfo.name
        self.msg = msg
        self.msg_key = msg_key

        for i in range(0, self.fieldInfo.count):
            if fieldInfo.bitfieldInfo == None:
                messageFieldTreeItem = FieldItem(self.editable, tree_widget, self.msg, self.msg_key, self.fieldInfo, [], i)
                self.addChild(messageFieldTreeItem)
            else:
                # This adds a bitfield container item, which in turn has bits within it.
                messageFieldTreeItem = FieldBitfieldItem(self.editable, tree_widget, self.msg, self.msg_key, fieldInfo, i)
                self.addChild(messageFieldTreeItem)

        if self.editable:
            self.setFlags(self.flags() | Qt.ItemIsEditable)

    def editableColumn(self, column):
        return False

    def data(self, column, role):
        if column != 2:
            return super(FieldArrayItem, self).data(column, role)

        # There's no data to display for the array itself, all data is in children.
        if role == Qt.FontRole:
            return QFont()
        if role == Qt.ForegroundRole:
            brush = QBrush()
            return brush
        if role == Qt.DisplayRole:
            if "hex" == self.fieldInfo.units.lower():
                ret = "0x"
                for i in range(0,self.fieldInfo.count):
                    ret += Messaging.get(self.msg, self.fieldInfo, i).replace("0x","")
                    if len(ret) > FieldArrayItem.CHILD_STRING_MAX_LENGTH:
                        ret += "..."
                        break
                return ret
            else:
                ret = ""
                for i in range(0,self.fieldInfo.count):
                    ret += str(Messaging.get(self.msg, self.fieldInfo, i))+", "
                    if len(ret) > FieldArrayItem.CHILD_STRING_MAX_LENGTH:
                        # remove the ", " from the string before appending ellipses,
                        # and then add two extra bytes at the end before returning
                        # because the return statements *also* removes two bytes.
                        ret = ret[:-2]+"..., "
                        break
                return ret[:-2]
                
            return ""
        return super(FieldArrayItem, self).data(column, role)

    def setData(self, column, role, value):
        # There's no data to set for the array, all data is in children.
        return

class QObjectProxy(QObject):
    send_message = pyqtSignal(object)
    def __init__(self):
        QObject.__init__(self)

class MessageItem(QTreeWidgetItem):
    def __init__(self, editable, tree_widget, msg, msg_key):
        QTreeWidgetItem.__init__(self, tree_widget, [msg.MsgName()])

        self.editable = editable
        
        self.qobjectProxy = QObjectProxy()

        self.tree_widget = tree_widget

        self.msg = msg
        self.msg_key = msg_key

        self.setup_fields(tree_widget)

        tree_widget.addTopLevelItem(self)
        self.setFirstColumnSpanned(True)
        self.setExpanded(True)

        # Create a timer to refresh at a fixed rate, so we don't refresh
        # needlessly fast for high rate data, and use too much CPU.
        # The timer isn't activated unless application code calls
        # set_msg_buffer().
        self.repaint_timer = QTimer()
        self.repaint_timer.setInterval(250)
        self.repaint_timer.timeout.connect(self.repaintAll)
        
        if self.editable:
            # text entry for rate, and a 'send' button
            self.sendTimer = QTimer()
            self.sendTimer.timeout.connect(lambda: self.qobjectProxy.send_message.emit(self.msg))
            containerWidget = QWidget(tree_widget)
            containerLayout = QHBoxLayout()
            containerWidget.setLayout(containerLayout)
            self.sendButton = QPushButton("Send", tree_widget)
            self.sendButton.autoFillBackground()
            self.sendButton.clicked.connect(self.sendClicked)
            self.rateEdit = QLineEdit("", tree_widget)
            self.rateEdit.setValidator(QDoubleValidator(0, 100, 2, containerWidget))
            self.rateEdit.setMaximumWidth(50)
            self.rateEdit.setFixedWidth(50)
            self.rateEdit.setPlaceholderText("Rate")
            # Set first column's text to empty, and replace it with a layout
            # with a few items in it, including the text we removed, rateEdit,
            # a "Hz" label, and a Send button.
            self.setText(0,"")
            containerLayout.addWidget(QLabel(msg.MsgName()))
            containerLayout.addWidget(self.rateEdit)
            containerLayout.addWidget(QLabel("Hz"))
            containerLayout.addWidget(self.sendButton)
            tree_widget.setItemWidget(self, 0, containerWidget)
            
        for i in range(0, tree_widget.columnCount()):
            tree_widget.resizeColumnToContents(i);

    def repaintAll(self):
        # Refresh the paint on the entire tree
        # TODO This is not a good solution!  We should refresh *only* the item that changed, not whole tree!
        region = self.tree_widget.childrenRegion()
        self.tree_widget.setDirtyRegion(region)
        self.repaint_timer.stop()

    def set_msg_buffer(self, msg_buffer):
        self.msg.msg_buffer_wrapper["msg_buffer"] = msg_buffer
        self.msg.hdr.msg_buffer_wrapper["msg_buffer"] = msg_buffer
        if not self.repaint_timer.isActive():
            self.repaint_timer.start()
        #self.repaintAll()

    def setup_fields(self, tree_widget):
        headerTreeItemParent = QTreeWidgetItem(self, [ "Header" ])
        self.addChild(headerTreeItemParent)
        headerTreeItemParent.setFirstColumnSpanned(True)

        for headerFieldInfo in Messaging.hdr.fields:
            if headerFieldInfo.bitfieldInfo != None:
                headerFieldTreeItem = FieldBitfieldItem(self.editable, tree_widget, self.msg.hdr, None, headerFieldInfo, index=None)
            else:
                headerFieldTreeItem = FieldItem(self.editable, tree_widget, self.msg.hdr, None, headerFieldInfo, column_strings=[], index=None)
            headerTreeItemParent.addChild(headerFieldTreeItem)

        for fieldInfo in type(self.msg).fields:
            messageFieldTreeItem = None

            if fieldInfo.count == 1:
                if fieldInfo.bitfieldInfo != None:
                    messageFieldTreeItem = FieldBitfieldItem(self.editable, tree_widget, self.msg, self.msg_key, fieldInfo, index=None)
                else:
                    messageFieldTreeItem = FieldItem(self.editable, tree_widget, self.msg, self.msg_key, fieldInfo, column_strings=[], index=None)
            else:
                messageFieldTreeItem = FieldArrayItem(self.editable, tree_widget, self.msg, self.msg_key, fieldInfo)
            
            self.addChild(messageFieldTreeItem)
            try:
                messageFieldTreeItem.overrideWidget
                tree_widget.setItemWidget(messageFieldTreeItem, 2, messageFieldTreeItem.overrideWidget)
            except AttributeError:
                pass

    def sendClicked(self):
        if self.sendButton.text() == "Send":
            if self.rateEdit.text() == "" or self.rateEdit.text() == "0":
                self.qobjectProxy.send_message.emit(self.msg)
            else:
                self.sendButton.setText("Stop")
                rate = float(self.rateEdit.text())
                if rate > 0.01:
                    period = int(1000.0/rate)
                    # send once now, then periodically after
                    self.qobjectProxy.send_message.emit(self.msg)
                    self.sendTimer.start(period)
                else:
                    self.sendButton.text("Send")
        else:
            self.sendButton.setText("Send")
            self.sendTimer.stop()



#------------------------------------------------------------------------------
# This reads a JSON log file into a dictionary that for each message name, has
# a list of all the messages of that type, with each message as dictionary with
# attribute access.
#------------------------------------------------------------------------------
import json
import collections
import types

# this allows getting keys in two ways:
# d['key']
# d.key
class DictWithAttributeAccess(dict):
    def __getattr__(self, key):
        return self[key] 
    def __setattr__(self, key, value):
        self[key] = value

def split(filename):
    # this is a dictionary of all the message types, and each
    # value is a list of messages of that type.
    msgs = DictWithAttributeAccess()

    with open(filename, 'r') as f:
        # read a line at a time, and append each message to a list for that message
        for line in f:
            j = json.loads(line, object_hook = lambda dict: DictWithAttributeAccess(dict))
            for msgname, msg in j.items():
                safename = msgname.replace('.','_')
                if not safename in msgs:
                    msgs[safename] = []
                msgs[safename].append(msg)
    return msgs



#!/usr/bin/env python3
import argparse
import json
import os
import sys
import time
from .pytimedinput import GetKey

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)

from msgtools.lib.file_reader import MessageFileReader
from msgtools.lib.messaging import Messaging
import msgtools.lib.msgcsv as msgcsv
import msgtools.lib.msgjson as msgjson
from msgtools.console.client import Client

DESCRIPTION='''
    msgreplay plays back recorded messages.
'''
class Replay(MessageFileReader):
    def __init__(self):
        super(Replay, self).__init__()

        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser.add_argument('logfile', help='''The log file you want to play back.  
            .bin or .log extension assumes the log is binary.
            .json assumes the log file is JSON.''')
        parser.add_argument('--speed', type=float, default="1.0", help='''Speed.  Less than 1 to slow down, greater than 1 to speed up.''')
        parser.add_argument('-i', '--include', nargs='+', help='''A list of message names to exclude from replay.''')
        parser.add_argument('-e', '--exclude', nargs='+', help='''A list of message names to exclude from replay.''')
        parser.add_argument('--debug', action='store_true', help='''Set to get more debug info.''')
        parser.add_argument('--ignoreinvalid', nargs='?', default=False, help='''Set to ignore invalid messages and fields in the log file.''')
        parser.add_argument('--newtime', action='store_true', help='''Set new timestamps in the messages we replay.  Default is to use original timestamps.''')
        self.args = parser.parse_args()

        # Open a client connection to replay the messages on
        self.connection = Client("Replay")

        # Figure out what to include/exclude
        self.excluded_msgs = None
        if self.args.exclude:
            if self.args.debug:
                print("Excluding %s" % (self.args.exclude))
            self.excluded_msgs = {}
            for msg_name in self.args.exclude:
                msg_id = self.msgid_from_name(msg_name)
                self.excluded_msgs[msg_id] = True
            if self.args.debug:
                print("Excluded: %s" % (list(self.excluded_msgs.keys())))
        self.included_msgs = None
        if self.args.include:
            if self.args.debug:
                print("Including %s" % (self.args.include))
            self.included_msgs = {}
            for msg_name in self.args.include:
                msg_id = self.msgid_from_name(msg_name)
                self.included_msgs[msg_id] = True
            if self.args.debug:
                print("Included: %s" % (list(self.included_msgs.keys())))
        self.message_sent_count = 0
        self.message_skipped_count = 0
        self.total_message_sent_count = 0
        self.total_message_skipped_count = 0
        
        time_units = Messaging.hdr.GetTime.units
        if time_units == "ms":
            self.time_scale = 1.0e-3
        elif time_units == "us":
            self.time_scale = 1.0e-6
        elif time_units == "ns":
            self.time_scale = 1.0e-9
        elif time_units == "s":
            self.time_scale = 1.0
        else:
            print("Invalid time units %s in Message header, exiting." % (time_units))
            exit(1)

        self.last_msg_time = None
        self.last_time = time.time()
        print("Press space or p to pause, f to speed up, s to slow down, ctrl-c to quit")
        try:
            if self.args.ignoreinvalid != None:
                # If it's not none, set the value the user passed in
                ignore_invalid = self.args.ignoreinvalid
            else:
                # If the arg is None, the user used the option, but didn't specify a value, so use true.
                ignore_invalid = True
            self.read_file(self.args.logfile, "NetworkHeader", ignore_invalid=ignore_invalid)
        except KeyError as e:
            print("\n\nKeyError!\nRe-run with --ignoreinvalid or --ignoreinvalid=silent if you want to ignore these!\n")
            raise e
        except KeyboardInterrupt:
            pass
        print("\nFinal Log time %.3f: replayed %d, skipped %d" % (self.last_msg_time, self.total_message_sent_count, self.total_message_skipped_count))

    def msgid_from_name(self, msg_name):
        try:
            msg_id = int(msg_name)
            return msg_id
        except:
            pass
        try:
            msg_class = Messaging.MsgClassFromName[msg_name]
            id = msg_class.ID
            return id
        except KeyError:
            print("ERROR!  Invalid message name " + msg_name)
            if self.args.debug:
                print("not in: %s" % ("\n".join(list(Messaging.MsgClassFromName.keys()))))
            quit()


    def process_message(self, msg):
        id = msg.hdr.GetMessageID()

        if self.excluded_msgs:
            if id in self.excluded_msgs:
                self.message_skipped_count += 1
                return
        if self.included_msgs:
            if not id in self.included_msgs:
                self.message_skipped_count += 1
                return
        
        msg_time = msg.hdr.GetTime() * self.time_scale
        if self.args.newtime:
            msg.hdr.SetTime(currenttime() / self.time_scale)
        
        def do_speeds(user_key):
            if user_key == 'f':
                self.args.speed *= 2
                print("  Speeding up to speed %s" % (self.args.speed))
            elif user_key == 's':
                self.args.speed /= 2
                print("  Slowing down to speed %s" % (self.args.speed))
            
        if self.last_msg_time == None:
            self.last_msg_time = msg_time
        else:
            delta_time = (msg_time - self.last_msg_time) / self.args.speed
            if delta_time <= 0.0:
                pass
            else:
                user_key = GetKey(allowCharacters=" sfp", timeout=delta_time)
                if user_key != None:
                    if user_key in [' ','p']:
                        while True:
                            print("  Paused!  Press space or p to un-pause, f to speed up, s to slow down, ctrl-c to quit")
                            user_key = GetKey(allowCharacters=" sfp", timeout=10)
                            if user_key != None:
                                do_speeds(user_key)
                                break
                    else:
                        do_speeds(user_key)
                self.last_msg_time = msg_time
        self.message_sent_count += 1
        self.connection.send(msg)
        current_time = time.time()
        if current_time > self.last_time + 1.0:
            self.last_time = current_time
            skipped_str = ", skipped %d" % self.message_skipped_count if self.message_skipped_count else ""
            print("Log time %.3f: Replayed %d%s" % (self.last_msg_time, self.message_sent_count, skipped_str))
            self.total_message_sent_count += self.message_sent_count
            self.total_message_skipped_count += self.message_skipped_count
            self.message_sent_count=0
            self.message_skipped_count=0

def main():
    replay = Replay()

# main starts here
if __name__ == '__main__':
    main()
